#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools
from collections import Iterable
from math import ceil, log

from ....compat import lrange, izip, irange, builtins, six
from ....config import options
from ...core import Tensor
from ..utils import check_out_param, validate_axis
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


class TensorReduction(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _is_cum(cls):
        return False

    def _call(self, a, out):
        a = astensor(a)
        if out is not None and not isinstance(out, Tensor):
            raise TypeError('out should be Tensor object, got {0} instead'.format(type(out)))

        axis = getattr(self, 'axis', None)
        keepdims = getattr(self, 'keepdims', None)

        if self._is_cum():
            if axis is None:
                a, axis = a.ravel(), 0
                setattr(self, 'axis', axis)
            shape = a.shape
        else:
            axis = lrange(len(a.shape)) if axis is None else axis
            if not isinstance(axis, Iterable):
                axis = (axis,)
            axis = set(axis)

            shape = tuple(s if i not in axis else 1 for i, s in enumerate(a.shape)
                          if keepdims or i not in axis)

        t = self.new_tensor([a], shape)

        if out is None:
            return t

        check_out_param(out, t, 'same_kind')
        out_shape, out_dtype = out.shape, out.dtype
        # if `out` is specified, use out's dtype and shape
        if out_shape != t.shape:
            if out.ndim > t.ndim:
                raise ValueError('output has too many dimensions')
            raise ValueError('output shape should be {0}, got {1}'.format(t.shape, out_shape))

        setattr(self, '_dtype', out_dtype)

        out.data = t.data
        return out

    def new_chunks(self, inputs, shape, **kw):
        chunks = super(TensorReduction, self).new_chunks(inputs, shape, **kw)
        setattr(self, '_input', getattr(self, '_inputs')[0])
        return chunks

    def new_tensors(self, inputs, shape, **kw):
        tensors = super(TensorReduction, self).new_tensors(inputs, shape, **kw)
        setattr(self, '_input', getattr(self, '_inputs')[0])
        return tensors

    def __call__(self, a, out=None):
        return self._call(a, out=out)

    @staticmethod
    def _reduced_shape(shape, axes):
        return tuple(1 if i in axes else s for i, s in enumerate(shape))

    @staticmethod
    def _reduced_nsplits(nsplits, axes):
        return tuple((1,) * len(c) if i in axes else c
                     for i, c in enumerate(nsplits))

    @staticmethod
    def _concatenate_shape(tensor, combine_block):
        return tuple(builtins.sum(nsplit[i] for i in cb)
                     for nsplit, cb in izip(tensor.nsplits, combine_block))

    @staticmethod
    def _combine_split(ax, combine_size, chunk_shape):
        if ax not in combine_size:
            return tuple((i,) for i in range(chunk_shape[ax]))
        else:
            size = combine_size[ax]
            shape = chunk_shape[ax]
            index = tuple(range(shape))
            return tuple(index[i:i + size] for i in irange(0, shape, size))

    def _get_op_kw(self):
        return None

    @staticmethod
    def _get_op_types():
        raise NotImplementedError

    @classmethod
    def _tree_reduction(cls, op, tensor, agg_op_type, axis, combine_op_type=None):
        kw = getattr(op, '_get_op_kw')() or {}
        keepdims = op.keepdims
        combine_size = op.combine_size or options.tensor.combine_size
        if isinstance(combine_size, dict):
            combine_size = dict((ax, combine_size.get(ax)) for ax in axis)
        else:
            assert isinstance(combine_size, six.integer_types)
            n = builtins.max(int(combine_size ** (1.0 / (len(axis) or 1))), 2)
            combine_size = dict((ax, n) for ax in axis)

        times = 1
        for i, n in enumerate(tensor.chunk_shape):
            if i in combine_size and combine_size[i] != 1:
                times = int(builtins.max(times, ceil(log(n, combine_size[i]))))

        for i in range(times - 1):
            combine_op = combine_op_type or agg_op_type
            [tensor] = cls._partial_reduction(combine_op, tensor, axis, op.dtype, True, combine_size)

        return cls._partial_reduction(agg_op_type, tensor, axis, op.dtype, keepdims, combine_size, kw)

    @classmethod
    def _partial_reduction(cls, agg_op_type, tensor, axis, dtype, keepdims, combine_size, kw=None):
        from ..merge.concatenate import TensorConcatenate
        kw = kw or {}
        axes = sorted(combine_size.keys())

        combine_blocks = [cls._combine_split(i, combine_size, tensor.chunk_shape)
                          for i in range(tensor.ndim)]
        combine_blocks_idxes = [range(len(blocks)) for blocks in combine_blocks]

        chunks = []
        for combine_block_idx, combine_block in izip(itertools.product(*combine_blocks_idxes),
                                                     itertools.product(*combine_blocks)):
            chks = [tensor.cix[idx] for idx in itertools.product(*combine_block)]
            op = TensorConcatenate(axis=axes, dtype=chks[0].dtype)
            chk = op.new_chunk(chks, shape=cls._concatenate_shape(tensor, combine_block))
            shape = tuple(s if i not in combine_size else 1
                          for i, s in enumerate(chk.shape) if keepdims or i not in combine_size)
            agg_op = agg_op_type(axis=axis, dtype=dtype, keepdims=keepdims, **kw)
            chunk = agg_op.new_chunk([chk], shape=shape,
                                     index=tuple(idx for i, idx in enumerate(combine_block_idx)
                                                 if keepdims or i not in combine_size))
            chunks.append(chunk)

        nsplits = [
            tuple(c.shape[i] for c in chunks if builtins.all(idx == 0 for j, idx in enumerate(c.index) if j != i))
            for i in range(len(chunks[0].shape))]
        shape = tuple(builtins.sum(nsplit) for nsplit in nsplits)
        agg_op = agg_op_type(axis=axis, dtype=dtype, keepdims=keepdims, combine_size=combine_size, **kw)
        return agg_op.new_tensors([tensor], shape, chunks=chunks, nsplits=nsplits)

    @classmethod
    def tile(cls, op):
        in_tensor = op.inputs[0]
        axis = tuple(range(in_tensor.ndim)) if op.axis is None else op.axis
        if isinstance(axis, six.integer_types):
            axis = (axis,)
        axis = tuple(validate_axis(in_tensor.ndim, ax) for ax in axis)

        if len(in_tensor.chunks) == 1:
            c = in_tensor.chunks[0]
            new_op = op.copy().reset_key()
            setattr(new_op, '_axis', axis)
            shape = list(cls._reduced_shape(c.shape, axis))
            nsplits = list(cls._reduced_nsplits(in_tensor.nsplits, axis))
            if not op.keepdims and axis:
                for ax in axis:
                    shape[ax] = None
                    nsplits[ax] = None
            shape = tuple(s for s in shape if s is not None)
            nsplits = tuple(ns for ns in nsplits if ns is not None)

            chunks = new_op.new_chunks([c], shape, index=c.index)
            return op.copy().new_tensors(op.inputs, op.outputs[0].shape, chunks=chunks, nsplits=nsplits)

        chunk_op_type, agg_op_type, combine_op_type = getattr(op, '_get_op_types')()

        chunks = []
        kw = getattr(op, '_get_op_kw')() or {}
        for c in in_tensor.chunks:
            chunk_op = chunk_op_type(axis=axis, dtype=op.dtype, keepdims=True,
                                     combine_size=op.combine_size, **kw)
            chunks.append(chunk_op.new_chunk([c], cls._reduced_shape(c.shape, axis), index=c.index))

        new_op = op.copy()
        tensor = new_op.new_tensor(op.inputs, cls._reduced_shape(in_tensor.shape, axis),
                                   nsplits=cls._reduced_nsplits(in_tensor.nsplits, axis), chunks=chunks)
        return cls._tree_reduction(new_op, tensor, agg_op_type, axis, combine_op_type=combine_op_type)


class TensorArgReduction(TensorReduction):
    __slots__ = ()

    @staticmethod
    def _get_arg_axis(axis, ndim):
        if axis is None:
            axis = tuple(range(ndim))
            ravel = True
        elif isinstance(axis, six.integer_types):
            if axis < 0:
                axis += ndim
            if axis < 0 or axis >= ndim:
                raise ValueError("axis entry is out of bounds")
            axis = (axis,)
            ravel = ndim == 1
        else:
            raise TypeError("axis must be either `None` or int, "
                            "got '{0}'".format(axis))
        return axis, ravel

    @staticmethod
    def _get_offset(tensor, axis, chunk, ravel):
        nsplits = tensor.nsplits
        offset = tuple(builtins.sum(split[:idx]) for split, idx in zip(nsplits, chunk.index))
        if not ravel:
            offset = offset[axis[0]]
        return offset

    @classmethod
    def tile(cls, op):
        in_tensor = op.inputs[0]
        axis, ravel = cls._get_arg_axis(op.axis, in_tensor.ndim)
        chunk_op_type, agg_op_type, combine_op_type = getattr(op, '_get_op_types')()

        chunks = []
        for c in in_tensor.chunks:
            offset = cls._get_offset(in_tensor, axis, c, ravel)
            chunk_op = chunk_op_type(axis=axis, dtype=op.dtype, keepdims=True,
                                     offset=offset, total_shape=in_tensor.shape,
                                     combine_size=op.combine_size)
            chunk = chunk_op.new_chunk([c], cls._reduced_shape(c.shape, axis), index=c.index)
            chunks.append(chunk)
        new_op = op.copy()
        tensor = new_op.new_tensor(op.inputs, cls._reduced_shape(in_tensor.shape, axis),
                                   nsplits=cls._reduced_nsplits(in_tensor.nsplits, axis), chunks=chunks)
        return cls._tree_reduction(new_op, tensor, agg_op_type, axis, combine_op_type=combine_op_type)


class TensorCumReduction(TensorReduction):
    __slots__ = ()

    @classmethod
    def _is_cum(cls):
        return True

    @staticmethod
    def _get_op_types():
        raise NotImplementedError

    @classmethod
    def tile(cls, op):
        from ..indexing.slice import TensorSlice

        in_tensor = op.inputs[0]
        axis = op.axis
        if not isinstance(axis, six.integer_types):
            raise ValueError("axis must be a integer")
        axis = validate_axis(in_tensor.ndim, axis)
        if axis is None:
            raise NotImplementedError

        op_type, binop_type = getattr(op, '_get_op_types')()

        chunks = []
        for c in in_tensor.chunks:
            chunk_op = op_type(axis=op.axis, dtype=op.dtype)
            chunks.append(chunk_op.new_chunk([c], c.shape, index=c.index))
        inter_tensor = copy.copy(in_tensor)
        inter_tensor._chunks = chunks

        slc = tuple(slice(None) if i != axis else slice(-1, None) for i in range(in_tensor.ndim))

        output_chunks = []
        for chunk in chunks:
            if chunk.index[axis] == 0:
                output_chunks.append(chunk)
                continue

            to_cum_chunks = [chunk]
            for i in range(chunk.index[axis]):
                to_cum_index = chunk.index[:axis] + (i,) + chunk.index[axis + 1:]
                shape = chunk.shape[:axis] + (1,) + chunk.shape[axis + 1:]
                to_cum_chunk = inter_tensor.cix[to_cum_index]
                slice_op = TensorSlice(slices=slc, dtype=chunk.dtype)
                sliced_chunk = slice_op.new_chunk([to_cum_chunk], shape, index=to_cum_index)
                to_cum_chunks.append(sliced_chunk)

            binop = binop_type(dtype=chunk.dtype)
            output_chunk = binop.new_chunk(to_cum_chunks, chunk.shape, index=chunk.index)
            output_chunks.append(output_chunk)

        nsplits = tuple((builtins.sum(c),) if i == axis else c for i, c in enumerate(in_tensor.nsplits))
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, in_tensor.shape, chunks=output_chunks, nsplits=nsplits)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fft import fft
from .ifft import ifft
from .fft2 import fft2
from .ifft2 import ifft2
from .fftn import fftn
from .ifftn import ifftn
from .rfft import rfft
from .irfft import irfft
from .rfft2 import rfft2
from .irfft2 import irfft2
from .rfftn import rfftn
from .irfftn import irfftn
from .hfft import hfft
from .ihfft import ihfft
from .fftfreq import fftfreq
from .rfftfreq import rfftfreq
from .fftshift import fftshift
from .ifftshift import ifftshift



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from numbers import Integral

import numpy as np

from ...core import TENSOR_TYPE
from ..utils import calc_sliced_size, broadcast_shape, replace_ellipsis, index_ndim
from ..datasource import tensor as astensor


_INDEX_ERROR_MSG = 'only integers, slices (`:`), ellipsis (`...`), ' \
                   'numpy.newaxis (`None`) and integer or boolean arrays are valid indices'


def get_index_and_shape(tensor, index):
    shape = []
    idx = 0
    fancy_index = None
    fancy_index_shapes = []
    for ind in index:
        if isinstance(ind, TENSOR_TYPE):
            if ind.dtype == np.bool_:
                shape.append(np.nan)
                for i, t_size, size in zip(itertools.count(0), ind.shape, tensor.shape[idx:ind.ndim + idx]):
                    if not np.isnan(t_size) and not np.isnan(size) and t_size != size:
                        raise IndexError(
                            'boolean index did not match indexed array along dimension {0}; '
                            'dimension is {1} but corresponding boolean dimension is {2}'.format(
                                idx + i, size, t_size)
                        )
                idx += ind.ndim
        elif isinstance(ind, np.ndarray):
            if fancy_index is None:
                fancy_index = idx
            fancy_index_shapes.append(ind.shape)
            idx += 1
        elif isinstance(ind, slice):
            if np.isnan(tensor.shape[idx]):
                shape.append(np.nan)
                idx += 1
            else:
                shape.append(calc_sliced_size(tensor.shape[idx], ind))
                idx += 1
        elif isinstance(ind, Integral):
            size = tensor.shape[idx]
            if not np.isnan(size) and ind >= size:
                raise IndexError('index {0} is out of bounds for axis {1} with size {2}'.format(
                    ind, idx, size
                ))
            idx += 1
        else:
            assert ind is None
            shape.append(1)

    if fancy_index is not None:
        try:
            fancy_index_shape = broadcast_shape(*fancy_index_shapes)
            shape = shape[:fancy_index] + list(fancy_index_shape) + shape[fancy_index:]
        except ValueError:
            raise IndexError(
                'shape mismatch: indexing arrays could not be broadcast together '
                'with shapes {0}'.format(' '.join(str(s) for s in fancy_index_shapes)))

    return index, shape


def preprocess_index(index):
    inds = []
    for ind in index:
        if isinstance(ind, (list, np.ndarray)):
            ind = np.array(ind)
            if ind.dtype.kind not in 'biu':
                raise IndexError(_INDEX_ERROR_MSG)
            if ind.dtype.kind == 'b':
                ind = astensor(ind)
        elif isinstance(ind, TENSOR_TYPE):
            if ind.dtype.kind not in 'biu':
                raise IndexError(_INDEX_ERROR_MSG)
        elif not isinstance(ind, (slice, Integral)) and ind is not None \
                and ind is not Ellipsis:
            raise IndexError(_INDEX_ERROR_MSG)
        inds.append(ind)

    return tuple(inds)


def process_index(tensor, item):
    if isinstance(item, list):
        arr = np.array(item)
        if arr.dtype == np.object_:
            item = tuple(item)
        elif arr.dtype.kind == 'f':
            raise IndexError(_INDEX_ERROR_MSG)
        else:
            item = (arr,)
    elif not isinstance(item, tuple):
        item = (item,)

    index = preprocess_index(item)
    index = replace_ellipsis(index, tensor.ndim)
    missing = tensor.ndim - sum(index_ndim(i) for i in index)
    if missing < 0:
        raise IndexError('too many indices for tensor')
    return index + (slice(None),) * missing



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import AnyField
from .core import TensorNoInput
from .array import tensor


class TensorFull(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_FULL

    _fill_value = AnyField('fill_value')

    def __init__(self, fill_value=None, dtype=None, gpu=None, **kw):
        if dtype is not None:
            dtype = np.dtype(dtype)
            if fill_value is not None:
                fill_value = dtype.type(fill_value)
        elif fill_value is not None:
            dtype = np.array(fill_value).dtype
        super(TensorFull, self).__init__(_fill_value=fill_value, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def fill_value(self):
        return self._fill_value


def full(shape, fill_value, dtype=None, chunk_size=None, gpu=False):
    """
    Return a new tensor of given shape and type, filled with `fill_value`.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the new tensor, e.g., ``(2, 3)`` or ``2``.
    fill_value : scalar
        Fill value.
    dtype : data-type, optional
        The desired data-type for the tensor  The default, `None`, means
         `np.array(fill_value).dtype`.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        Tensor of `fill_value` with the given shape, dtype, and order.

    See Also
    --------
    zeros_like : Return a tensor of zeros with shape and type of input.
    ones_like : Return a tensor of ones with shape and type of input.
    empty_like : Return an empty tensor with shape and type of input.
    full_like : Fill a tensor with shape and type of input.
    zeros : Return a new tensor setting values to zero.
    ones : Return a new tensor setting values to one.
    empty : Return a new uninitialized tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.full((2, 2), mt.inf).execute()
    array([[ inf,  inf],
           [ inf,  inf]])
    >>> mt.full((2, 2), 10).execute()
    array([[10, 10],
           [10, 10]])

    """
    v = np.asarray(fill_value)
    if len(v.shape) > 0:
        from ..base import broadcast_to
        return broadcast_to(tensor(v, dtype=dtype, chunk_size=chunk_size, gpu=gpu), shape)

    op = TensorFull(fill_value, dtype=dtype, gpu=gpu)
    return op(shape, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Operand
from ..core import BaseWithKey
from .. import opcodes as OperandDef
from ..serialize import ValueType, AnyField, KeyField, TupleField, \
    StringField, Int32Field, BoolField, DictField, Float64Field


class ElementWise(Operand):
    __slots__ = ()


class Constant(ElementWise):
    _lhs = AnyField('lhs')
    _rhs = AnyField('rhs')
    _casting = StringField('casting')
    _err = DictField('err', ValueType.string, ValueType.string)

    @property
    def lhs(self):
        return self._lhs

    @property
    def rhs(self):
        return self._rhs

    @property
    def constant(self):
        if isinstance(self._lhs, BaseWithKey):
            return [self._rhs]
        elif isinstance(self._rhs, BaseWithKey):
            return [self._lhs]
        return [self._lhs, self._rhs]

    @property
    def reverse(self):
        return isinstance(self._rhs, BaseWithKey)

    @property
    def casting(self):
        return getattr(self, '_casting', None)

    @property
    def err(self):
        return getattr(self, '_err', dict())


class BinOp(ElementWise):
    _lhs = KeyField('lhs')
    _rhs = KeyField('rhs')
    _out = KeyField('out')
    _where = KeyField('where')
    _casting = StringField('casting')
    _err = DictField('err', ValueType.string, ValueType.string)

    @property
    def lhs(self):
        return self._lhs

    @property
    def rhs(self):
        return self._rhs

    @property
    def out(self):
        return getattr(self, '_out', None)

    @property
    def where(self):
        return getattr(self, '_where', None)

    @property
    def casting(self):
        return getattr(self, '_casting', None)

    @property
    def err(self):
        return getattr(self, '_err', dict())


class UnaryOp(ElementWise):
    _input = KeyField('input')
    _out = KeyField('out')
    _where = KeyField('where')
    _casting = StringField('casting')
    _err = DictField('err', ValueType.string, ValueType.string)

    @property
    def input(self):
        return self._input

    @property
    def out(self):
        return getattr(self, '_out', None)

    @property
    def where(self):
        return getattr(self, '_where', None)

    @property
    def casting(self):
        return getattr(self, '_casting', None)

    @property
    def err(self):
        return getattr(self, '_err', dict())


class Add(BinOp):
    _op_type_ = OperandDef.ADD


class AddConstant(Constant):
    _op_type_ = OperandDef.ADD_CONSTANT


class Subtract(BinOp):
    _op_type_ = OperandDef.SUB


class SubConstant(Constant):
    _op_type_ = OperandDef.SUB_CONSTANT


class Multiply(BinOp):
    _op_type_ = OperandDef.MUL


class MulConstant(Constant):
    _op_type_ = OperandDef.MUL_CONSTANT


class Divide(BinOp):
    _op_type_ = OperandDef.DIV


class DivConstant(Constant):
    _op_type_ = OperandDef.DIV_CONSTANT


class TrueDiv(BinOp):
    _op_type_ = OperandDef.TRUEDIV


class TDivConstant(Constant):
    _op_type_ = OperandDef.TDIV_CONSTANT


class FloorDiv(BinOp):
    _op_type_ = OperandDef.FLOORDIV


class FDivConstant(Constant):
    _op_type_ = OperandDef.FDIV_CONSTANT


class Power(BinOp):
    _op_type_ = OperandDef.POW


class PowConstant(Constant):
    _op_type_ = OperandDef.POW_CONSTANT


class FloatPower(BinOp):
    _op_type_ = OperandDef.FLOAT_POWER


class FloatPowerConstant(Constant):
    _op_type_ = OperandDef.FLOAT_POWER_CONSTANT


class Mod(BinOp):
    _op_type_ = OperandDef.MOD


class ModConstant(Constant):
    _op_type_ = OperandDef.MOD_CONSTANT


class FMod(BinOp):
    _op_type_ = OperandDef.FMOD


class FModConstant(Constant):
    _op_type_ = OperandDef.FMOD_CONSTANT


class LogAddExp(BinOp):
    _op_type_ = OperandDef.LOGADDEXP


class LAEConstant(Constant):
    _op_type_ = OperandDef.LAE_CONSTANT


class LogAddExp2(BinOp):
    _op_type_ = OperandDef.LOGADDEXP2


class LAE2Constant(Constant):
    _op_type_ = OperandDef.LAE2_CONSTANT


class Isclose(BinOp):
    _op_type_ = OperandDef.ISCLOSE

    _rtol = Float64Field('rtol')
    _atol = Float64Field('atol')
    _equal_nan = BoolField('equal_nan')

    @property
    def rtol(self):
        return self._rtol

    @property
    def atol(self):
        return self._atol

    @property
    def equal_nan(self):
        return self._equal_nan


class IscloseConstant(Constant):
    _op_type_ = OperandDef.ISCLOSE_CONSTANT

    _rtol = Float64Field('rtol')
    _atol = Float64Field('atol')
    _equal_nan = BoolField('equal_nan')

    @property
    def rtol(self):
        return self._rtol

    @property
    def atol(self):
        return self._atol

    @property
    def equal_nan(self):
        return self._equal_nan


class Negative(UnaryOp):
    _op_type_ = OperandDef.NEGATIVE


class Positive(UnaryOp):
    _op_type_ = OperandDef.POSITIVE


class Absolute(UnaryOp):
    _op_type_ = OperandDef.ABSOLUTE


class Fabs(UnaryOp):
    _op_type_ = OperandDef.FABS


class Abs(UnaryOp):
    _op_type_ = OperandDef.ABS


class Rint(UnaryOp):
    _op_type_ = OperandDef.RINT


class Sign(UnaryOp):
    _op_type_ = OperandDef.SIGN


class Degrees(UnaryOp):
    _op_type_ = OperandDef.DEGREES


class Radians(UnaryOp):
    _op_type_ = OperandDef.RADIANS


class Conj(UnaryOp):
    _op_type_ = OperandDef.CONJ


class Exp(UnaryOp):
    _op_type_ = OperandDef.EXP


class Exp2(UnaryOp):
    _op_type_ = OperandDef.EXP2


class Log(UnaryOp):
    _op_type_ = OperandDef.LOG


class Log2(UnaryOp):
    _op_type_ = OperandDef.LOG2


class Log10(UnaryOp):
    _op_type_ = OperandDef.LOG10


class Expm1(UnaryOp):
    _op_type_ = OperandDef.EXPM1


class Log1p(UnaryOp):
    _op_type_ = OperandDef.LOG1P


class Sqrt(UnaryOp):
    _op_type_ = OperandDef.SQRT


class Square(UnaryOp):
    _op_type_ = OperandDef.SQUARE


class Cbrt(UnaryOp):
    _op_type_ = OperandDef.CBRT


class Around(UnaryOp):
    _op_type_ = OperandDef.AROUND

    _decimals = Int32Field('decimals')

    @property
    def decimals(self):
        return self._decimals


class Reciprocal(UnaryOp):
    _op_type_ = OperandDef.RECIPROCAL


class Equal(BinOp):
    _op_type_ = OperandDef.EQ


class EqConstant(Constant):
    _op_type_ = OperandDef.EQ_CONSTANT


class NotEqual(BinOp):
    _op_type_ = OperandDef.NE


class NeConstant(Constant):
    _op_type_ = OperandDef.NE_CONSTANT


class LessThan(BinOp):
    _op_type_ = OperandDef.LT


class LtConstant(Constant):
    _op_type_ = OperandDef.LT_CONSTANT


class LessEqual(BinOp):
    _op_type_ = OperandDef.LE


class LeConstant(Constant):
    _op_type_ = OperandDef.LE_CONSTANT


class GreaterThan(BinOp):
    _op_type_ = OperandDef.GT


class GtConstant(Constant):
    _op_type_ = OperandDef.GT_CONSTANT


class GreaterEqual(BinOp):
    _op_type_ = OperandDef.GE


class GeConstant(Constant):
    _op_type_ = OperandDef.GE_CONSTANT


class Sin(UnaryOp):
    _op_type_ = OperandDef.SIN


class Cos(UnaryOp):
    _op_type_ = OperandDef.COS


class Tan(UnaryOp):
    _op_type_ = OperandDef.TAN


class Arcsin(UnaryOp):
    _op_type_ = OperandDef.ARCSIN


class Arccos(UnaryOp):
    _op_type_ = OperandDef.ARCCOS


class Arctan(UnaryOp):
    _op_type_ = OperandDef.ARCTAN


class Arctan2(BinOp):
    _op_type_ = OperandDef.ARCTAN2


class Arct2Constant(Constant):
    _op_type_ = OperandDef.ARCT2_CONSTANT


class Hypot(BinOp):
    _op_type_ = OperandDef.HYPOT


class HypotConstant(Constant):
    _op_type_ = OperandDef.HYPOT_CONSTANT


class Sinh(UnaryOp):
    _op_type_ = OperandDef.SINH


class Cosh(UnaryOp):
    _op_type_ = OperandDef.COSH


class Tanh(UnaryOp):
    _op_type_ = OperandDef.TANH


class Arcsinh(UnaryOp):
    _op_type_ = OperandDef.ARCSINH


class Arccosh(UnaryOp):
    _op_type_ = OperandDef.ARCCOSH


class Arctanh(UnaryOp):
    _op_type_ = OperandDef.ARCTANH


class Deg2rad(UnaryOp):
    _op_type_ = OperandDef.DEG2RAD


class Rad2deg(UnaryOp):
    _op_type_ = OperandDef.RAD2DEG


class Bitand(BinOp):
    _op_type_ = OperandDef.BITAND


class BitandConstant(Constant):
    _op_type_ = OperandDef.BITAND_CONSTANT


class Bitor(BinOp):
    _op_type_ = OperandDef.BITOR


class BitorConstant(Constant):
    _op_type_ = OperandDef.BITOR_CONSTANT


class Bitxor(BinOp):
    _op_type_ = OperandDef.BITXOR


class BitxorConstant(Constant):
    _op_type_ = OperandDef.BITXOR_CONSTANT


class Invert(UnaryOp):
    _op_type_ = OperandDef.INVERT


class Lshift(BinOp):
    _op_type_ = OperandDef.LSHIFT


class LshiftConstant(Constant):
    _op_type_ = OperandDef.LSHIFT_CONSTANT


class Rshift(BinOp):
    _op_type_ = OperandDef.RSHIFT


class RshiftConstant(Constant):
    _op_type_ = OperandDef.RSHIFT_CONSTANT


class And(BinOp):
    _op_type_ = OperandDef.AND


class AndConstant(Constant):
    _op_type_ = OperandDef.AND_CONSTANT


class Or(BinOp):
    _op_type_ = OperandDef.OR


class OrConstant(Constant):
    _op_type_ = OperandDef.OR_CONSTANT


class Xor(BinOp):
    _op_type_ = OperandDef.XOR


class XorConstant(Constant):
    _op_type_ = OperandDef.XOR_CONSTANT


class Not(UnaryOp):
    _op_type_ = OperandDef.NOT


class Maximum(BinOp):
    _op_type_ = OperandDef.MAXIMUM


class MaximumConstant(Constant):
    _op_type_ = OperandDef.MAXIMUM_CONSTANT


class Minimum(BinOp):
    _op_type_ = OperandDef.MINIMUM


class MinimumConstant(Constant):
    _op_type_ = OperandDef.MINIMUM_CONSTANT


class Floor(UnaryOp):
    _op_type_ = OperandDef.FLOOR


class Ceil(UnaryOp):
    _op_type_ = OperandDef.CEIL


class Trunc(UnaryOp):
    _op_type_ = OperandDef.TRUNC


class FMax(BinOp):
    _op_type_ = OperandDef.FMAX


class FMaxConstant(Constant):
    _op_type_ = OperandDef.FMAX_CONSTANT


class FMin(BinOp):
    _op_type_ = OperandDef.FMIN


class FMinConstant(Constant):
    _op_type_ = OperandDef.FMIN_CONSTANT


class IsFinite(UnaryOp):
    _op_type_ = OperandDef.ISFINITE


class IsInf(UnaryOp):
    _op_type_ = OperandDef.ISINF


class IsNan(UnaryOp):
    _op_type_ = OperandDef.ISNAN


class Signbit(UnaryOp):
    _op_type_ = OperandDef.SIGNBIT


class Copysign(BinOp):
    _op_type_ = OperandDef.COPYSIGN


class CopysignConstant(Constant):
    _op_type_ = OperandDef.COPYSIGN_CONSTANT


class Nextafter(BinOp):
    _op_type_ = OperandDef.NEXTAFTER


class NextafterConstant(Constant):
    _op_type_ = OperandDef.NEXTAFTER_CONSTANT


class Spacing(UnaryOp):
    _op_type_ = OperandDef.SPACING


class Modf(ElementWise):
    _op_type_ = OperandDef.MODF

    _input = KeyField('input')
    _out1 = KeyField('out1')
    _out2 = KeyField('out2')
    _where = KeyField('where')
    _casting = StringField('casting')

    @property
    def output_limit(self):
        return 2

    @property
    def input(self):
        return self._input

    @property
    def out1(self):
        return getattr(self, '_out1', None)

    @property
    def out2(self):
        return getattr(self, '_out2', None)

    @property
    def where(self):
        return getattr(self, '_where', None)

    @property
    def casting(self):
        return getattr(self, '_casting', None)


class Ldexp(BinOp):
    _op_type_ = OperandDef.LDEXP


class LdexpConstant(Constant):
    _op_type_ = OperandDef.LDEXP_CONSTANT


class Frexp(ElementWise):
    _op_type_ = OperandDef.FREXP

    _input = KeyField('input')
    _out1 = KeyField('out1')
    _out2 = KeyField('out2')
    _where = KeyField('where')
    _casting = StringField('casting')

    @property
    def output_limit(self):
        return 2

    @property
    def input(self):
        return self._input

    @property
    def out1(self):
        return getattr(self, '_out1', None)

    @property
    def out2(self):
        return getattr(self, '_out2', None)

    @property
    def where(self):
        return getattr(self, '_where', None)

    @property
    def casting(self):
        return getattr(self, '_casting', None)


class Clip(ElementWise):
    _op_type_ = OperandDef.CLIP

    _a = KeyField('a')
    _a_min = AnyField('a_min')
    _a_max = AnyField('a_max')
    _out = KeyField('out')

    @property
    def a(self):
        return self._a

    @property
    def a_min(self):
        return self._a_min

    @property
    def a_max(self):
        return self._a_max

    @property
    def out(self):
        return getattr(self, '_out', None)


class Angle(UnaryOp):
    _op_type_ = OperandDef.ANGLE

    _deg = BoolField('deg')

    @property
    def deg(self):
        return self._deg


class IsReal(UnaryOp):
    _op_type_ = OperandDef.ISREAL


class IsComplex(UnaryOp):
    _op_type_ = OperandDef.ISCOMPLEX


class Real(UnaryOp):
    _op_type_ = OperandDef.REAL


class Imag(UnaryOp):
    _op_type_ = OperandDef.IMAG


class SetReal(BinOp):
    _op_type_ = OperandDef.SET_REAL


class SetRealConstant(Constant):
    _op_type_ = OperandDef.SET_REAL_CONSTANT


class SetImag(BinOp):
    _op_type_ = OperandDef.SET_IMAG


class SetImagConstant(Constant):
    _op_type_ = OperandDef.SET_IMAG_CONSTANT


class Fix(UnaryOp):
    _op_type_ = OperandDef.FIX


class I0(UnaryOp):
    _op_type_ = OperandDef.I0


class Sinc(UnaryOp):
    _op_type_ = OperandDef.SINC


class NanToNum(UnaryOp):
    _op_type_ = OperandDef.NAN_TO_NUM


class TreeAdd(ElementWise):
    _op_type_ = OperandDef.TREE_ADD


class TreeMultiply(ElementWise):
    _op_type_ = OperandDef.TREE_MULTIPLY


class TensorDot(Operand):
    _op_type_ = OperandDef.TENSORDOT

    _a = KeyField('a')
    _b = KeyField('b')
    _a_axes = TupleField('a_axes', ValueType.int32)
    _b_axes = TupleField('b_axes', ValueType.int32)

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def a_axes(self):
        return self._a_axes

    @property
    def b_axes(self):
        return self._b_axes


class Dot(Operand):
    _op_type_ = OperandDef.DOT

    _a = KeyField('a')
    _b = KeyField('b')

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b


class Matmul(Operand):
    _op_type_ = OperandDef.MATMUL

    _a = KeyField('a')
    _b = KeyField('b')

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from .... import operands
from ...core import ExecutableTuple
from ..datasource import tensor as astensor
from .core import TSQR


class TensorSVD(operands.SVD, TSQR):
    def __init__(self, method=None, dtype=None, **kw):
        super(TensorSVD, self).__init__(_method=method, _dtype=dtype, **kw)

    @classmethod
    def _is_svd(cls):
        return True

    def _set_inputs(self, inputs):
        super(TensorSVD, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        a = astensor(a)

        if a.ndim != 2:
            raise LinAlgError('{0}-dimensional tensor given. '
                              'Tensor must be two-dimensional'.format(a.ndim))

        tiny_U, tiny_s, tiny_V = np.linalg.svd(np.ones((1, 1), dtype=a.dtype))

        U_shape = a.shape
        s_shape = (a.shape[1],)
        V_shape = (a.shape[1],) * 2
        U, s, V = self.new_tensors([a], (U_shape, s_shape, V_shape),
                                   kws=[
                                       {'side': 'U', 'dtype': tiny_U.dtype},
                                       {'side': 's', 'dtype': tiny_s.dtype},
                                       {'side': 'V', 'dtype': tiny_V.dtype}
                                   ])
        return ExecutableTuple([U, s, V])

    @classmethod
    def tile(cls, op):
        U, s, V = op.outputs
        U_dtype, s_dtype, V_dtype = U.dtype, s.dtype, V.dtype
        U_shape, s_shape, V_shape = U.shape, s.shape, V.shape
        in_tensor = op.input
        if in_tensor.chunk_shape == (1, 1):
            in_chunk = in_tensor.chunks[0]
            chunk_op = op.copy().reset_key()
            svd_chunks = chunk_op.new_chunks([in_chunk], (U_shape, s_shape, V_shape),
                                             kws=[
                                                 {'side': 'U', 'dtype': U_dtype,
                                                  'index': in_chunk.index},
                                                 {'side': 's', 'dtype': s_dtype,
                                                  'index': in_chunk.index[1:]},
                                                 {'side': 'V', 'dtype': V_dtype,
                                                  'index': in_chunk.index}
                                             ])
            U_chunk, s_chunk, V_chunk = svd_chunks

            new_op = op.copy()
            kws = [
                {'chunks': [U_chunk], 'nsplits': tuple((s,) for s in U_shape), 'dtype': U_dtype},
                {'chunks': [s_chunk], 'nsplits': tuple((s,) for s in s_shape), 'dtype': s_dtype},
                {'chunks': [V_chunk], 'nsplits': tuple((s,) for s in V_shape), 'dtype': V_dtype}
            ]
            return new_op.new_tensors(op.inputs, [U_shape, s_shape, V_shape], kws=kws)
        elif op.method == 'tsqr':
            return super(TensorSVD, cls).tile(op)
        else:
            raise NotImplementedError('Only tsqr method supported for now')


def svd(a, method='tsqr'):
    """
    Singular Value Decomposition.

    When `a` is a 2D tensor, it is factorized as ``u @ np.diag(s) @ vh
    = (u * s) @ vh``, where `u` and `vh` are 2D unitary tensors and `s` is a 1D
    tensor of `a`'s singular values. When `a` is higher-dimensional, SVD is
    applied in stacked mode as explained below.

    Parameters
    ----------
    a : (..., M, N) array_like
        A real or complex tensor with ``a.ndim >= 2``.
    method: {'tsqr'}, optional
        method to calculate qr factorization, tsqr as default

        TSQR is presented in:

            A. Benson, D. Gleich, and J. Demmel.
            Direct QR factorizations for tall-and-skinny matrices in
            MapReduce architectures.
            IEEE International Conference on Big Data, 2013.
            http://arxiv.org/abs/1301.1071


    Returns
    -------
    u : { (..., M, M), (..., M, K) } tensor
        Unitary tensor(s). The first ``a.ndim - 2`` dimensions have the same
        size as those of the input `a`. The size of the last two dimensions
        depends on the value of `full_matrices`. Only returned when
        `compute_uv` is True.
    s : (..., K) tensor
        Vector(s) with the singular values, within each vector sorted in
        descending order. The first ``a.ndim - 2`` dimensions have the same
        size as those of the input `a`.
    vh : { (..., N, N), (..., K, N) } tensor
        Unitary tensor(s). The first ``a.ndim - 2`` dimensions have the same
        size as those of the input `a`. The size of the last two dimensions
        depends on the value of `full_matrices`. Only returned when
        `compute_uv` is True.

    Raises
    ------
    LinAlgError
        If SVD computation does not converge.

    Notes
    -----

    SVD is usually described for the factorization of a 2D matrix :math:`A`.
    The higher-dimensional case will be discussed below. In the 2D case, SVD is
    written as :math:`A = U S V^H`, where :math:`A = a`, :math:`U= u`,
    :math:`S= \\mathtt{np.diag}(s)` and :math:`V^H = vh`. The 1D tensor `s`
    contains the singular values of `a` and `u` and `vh` are unitary. The rows
    of `vh` are the eigenvectors of :math:`A^H A` and the columns of `u` are
    the eigenvectors of :math:`A A^H`. In both cases the corresponding
    (possibly non-zero) eigenvalues are given by ``s**2``.

    If `a` has more than two dimensions, then broadcasting rules apply, as
    explained in :ref:`routines.linalg-broadcasting`. This means that SVD is
    working in "stacked" mode: it iterates over all indices of the first
    ``a.ndim - 2`` dimensions and for each combination SVD is applied to the
    last two indices. The matrix `a` can be reconstructed from the
    decomposition with either ``(u * s[..., None, :]) @ vh`` or
    ``u @ (s[..., None] * vh)``. (The ``@`` operator can be replaced by the
    function ``mt.matmul`` for python versions below 3.5.)

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.random.randn(9, 6) + 1j*mt.random.randn(9, 6)
    >>> b = mt.random.randn(2, 7, 8, 3) + 1j*mt.random.randn(2, 7, 8, 3)

    Reconstruction based on reduced SVD, 2D case:

    >>> u, s, vh = mt.linalg.svd(a)
    >>> u.shape, s.shape, vh.shape
    ((9, 6), (6,), (6, 6))
    >>> np.allclose(a, np.dot(u * s, vh))
    True
    >>> smat = np.diag(s)
    >>> np.allclose(a, np.dot(u, np.dot(smat, vh)))
    True

    """
    op = TensorSVD(method=method)
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ....operands import Argwhere
from ..utils import recursive_tile
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor
from .ravel import ravel


class TensorArgwhere(Argwhere, TensorOperandMixin):
    def __init__(self, dtype=None, **kw):
        super(TensorArgwhere, self).__init__(_dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorArgwhere, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        shape = (np.nan, a.ndim)
        return self.new_tensor([a], shape)

    @classmethod
    def tile(cls, op):
        from ..datasource import arange
        from ..indexing import unravel_index
        from ..reshape.reshape import TensorReshape

        in_tensor = op.input

        flattened = ravel(in_tensor).single_tiles()
        indices = arange(flattened.size, dtype=np.intp, chunks=flattened.nsplits)
        indices = indices[flattened]
        dim_indices = unravel_index(indices, in_tensor.shape)
        [recursive_tile(ind) for ind in dim_indices]

        out_chunk_shape = dim_indices[0].chunk_shape + (in_tensor.ndim,)
        nsplits = dim_indices[0].nsplits + ((1,) * in_tensor.ndim,)
        out_chunks = []
        for out_index in itertools.product(*(map(range, out_chunk_shape))):
            dim_ind_chunk = dim_indices[out_index[1]].chunks[out_index[0]]
            chunk_shape = dim_ind_chunk.shape + (1,)
            chunk_op = TensorReshape(newshape=(-1, 1), dtype=dim_ind_chunk.dtype)
            out_chunk = chunk_op.new_chunk([dim_ind_chunk], chunk_shape, index=out_index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  chunks=out_chunks, nsplits=nsplits)


def argwhere(a):
    """
    Find the indices of tensor elements that are non-zero, grouped by element.

    Parameters
    ----------
    a : array_like
        Input data.

    Returns
    -------
    index_tensor : Tensor
        Indices of elements that are non-zero. Indices are grouped by element.

    See Also
    --------
    where, nonzero

    Notes
    -----
    ``mt.argwhere(a)`` is the same as ``mt.transpose(mt.nonzero(a))``.

    The output of ``argwhere`` is not suitable for indexing tensors.
    For this purpose use ``nonzero(a)`` instead.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(6).reshape(2,3)
    >>> x.execute()
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> mt.argwhere(x>1).execute()
    array([[0, 2],
           [1, 0],
           [1, 1],
           [1, 2]])

    """
    a = astensor(a).astype(bool)
    op = TensorArgwhere(np.intp)
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorNoncentralF(operands.NoncentralF, TensorRandomOperandMixin):
    __slots__ = '_dfnum', '_dfden', '_nonc', '_size'
    _input_fields_ = ['_dfnum', '_dfden', '_nonc']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorNoncentralF, self).__init__(_size=size, _state=state, _dtype=dtype,
                                                _gpu=gpu, **kw)

    def __call__(self, dfnum, dfden, nonc, chunk_size=None):
        return self.new_tensor([dfnum, dfden, nonc], None, raw_chunk_size=chunk_size)


def noncentral_f(random_state, dfnum, dfden, nonc, size=None, chunk_size=None, gpu=None, **kw):
    """
    Draw samples from the noncentral F distribution.

    Samples are drawn from an F distribution with specified parameters,
    `dfnum` (degrees of freedom in numerator) and `dfden` (degrees of
    freedom in denominator), where both parameters > 1.
    `nonc` is the non-centrality parameter.

    Parameters
    ----------
    dfnum : float or array_like of floats
        Numerator degrees of freedom, should be > 0.
    dfden : float or array_like of floats
        Denominator degrees of freedom, should be > 0.
    nonc : float or array_like of floats
        Non-centrality parameter, the sum of the squares of the numerator
        means, should be >= 0.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``dfnum``, ``dfden``, and ``nonc``
        are all scalars.  Otherwise, ``np.broadcast(dfnum, dfden, nonc).size``
        samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized noncentral Fisher distribution.

    Notes
    -----
    When calculating the power of an experiment (power = probability of
    rejecting the null hypothesis when a specific alternative is true) the
    non-central F statistic becomes important.  When the null hypothesis is
    true, the F statistic follows a central F distribution. When the null
    hypothesis is not true, then it follows a non-central F statistic.

    References
    ----------
    .. [1] Weisstein, Eric W. "Noncentral F-Distribution."
           From MathWorld--A Wolfram Web Resource.
           http://mathworld.wolfram.com/NoncentralF-Distribution.html
    .. [2] Wikipedia, "Noncentral F-distribution",
           http://en.wikipedia.org/wiki/Noncentral_F-distribution

    Examples
    --------
    In a study, testing for a specific alternative to the null hypothesis
    requires use of the Noncentral F distribution. We need to calculate the
    area in the tail of the distribution that exceeds the value of the F
    distribution for the null hypothesis.  We'll plot the two probability
    distributions for comparison.

    >>> import mars.tensor as mt
    >>> import matplotlib.pyplot as plt

    >>> dfnum = 3 # between group deg of freedom
    >>> dfden = 20 # within groups degrees of freedom
    >>> nonc = 3.0
    >>> nc_vals = mt.random.noncentral_f(dfnum, dfden, nonc, 1000000)
    >>> NF = np.histogram(nc_vals.execute(), bins=50, normed=True)  # TODO(jisheng): implement mt.histogram
    >>> c_vals = mt.random.f(dfnum, dfden, 1000000)
    >>> F = np.histogram(c_vals.execute(), bins=50, normed=True)
    >>> plt.plot(F[1][1:], F[0])
    >>> plt.plot(NF[1][1:], NF[0])
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().noncentral_f(
            handle_array(dfnum), handle_array(dfden), handle_array(nonc), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorNoncentralF(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(dfnum, dfden, nonc, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

from .graph import DirectedGraph


class Tilesable(object):
    __slots__ = ()

    def is_coarse(self):
        raise NotImplementedError

    def copy_from(self, other):
        raise NotImplementedError

    def to_graph(self, graph=None, graph_cls=DirectedGraph, substitutes=None):
        visited = set()
        graph = graph_cls() if graph is None else graph
        substitutes = substitutes or dict()

        if self in substitutes:
            graph.add_node(substitutes[self])
            return graph

        q = [self]
        while len(q) > 0:
            obj = q.pop()
            if obj in visited:
                continue

            if obj not in graph:
                graph.add_node(obj)
            for input in obj.inputs or []:
                sub = False
                if input in substitutes:
                    input = substitutes[input]
                    sub = True
                in_graph = True
                if input not in graph:
                    graph.add_node(input)
                    in_graph = False
                graph.add_edge(input, obj)
                if not sub and not in_graph:
                    q.append(input)

            visited.add(obj)

        return graph

    def _repr_svg_(self):
        return self.to_graph()._repr_svg_()


class TilesError(Exception):
    pass


class DataNotReady(TilesError):
    pass


class NotSupportTile(Exception):
    pass


class OperandTilesHandler(object):
    def __init__(self):
        self._handlers = {}

    @classmethod
    def _get_op_cls(cls, op):
        if isinstance(op, type):
            return op
        return type(op)

    @classmethod
    def _assign_to(cls, tile_after_tensor_datas, tile_before_tensor_datas):
        assert len(tile_after_tensor_datas) == len(tile_before_tensor_datas)

        for tile_after_tensor_data, tile_before_tensor_data in \
                zip(tile_after_tensor_datas, tile_before_tensor_datas):
            if tile_before_tensor_data is None:
                # garbage collected
                continue
            tile_after_tensor_data.copy_to(tile_before_tensor_data)
            tile_before_tensor_data.op.outputs = tile_before_tensor_datas

    def register(self, op, handler):
        self._handlers[self._get_op_cls(op)] = handler

    def _dispatch(self, op):
        op_cls = self._get_op_cls(op)
        try:
            handler = self._handlers[op_cls]
            return handler(op)
        except KeyError as e:
            if hasattr(op_cls, 'tile'):
                # has tile implementation
                return op_cls.tile(op)
            for op_clz in self._handlers.keys():
                if issubclass(op_cls, op_clz):
                    self._handlers[op_cls] = self._handlers[op_clz]
                    return self._handlers[op_cls](op)

            raise e

    def dispatch(self, to_tiles):
        return self._dispatch(to_tiles.op)

    def single_tiles(self, to_tiles):
        if to_tiles.is_coarse() and to_tiles.op:
            dispatched = self._dispatch(to_tiles.op)
            self._assign_to([d.data for d in dispatched], to_tiles.op.outputs)

        return to_tiles

    def tiles(self, tiles_obj):
        graph = DirectedGraph()
        visited = set([id(tiles_obj)])
        loose_requires = set()
        q = deque([tiles_obj])

        while q:
            to_tiles = q.popleft()
            if to_tiles not in graph:
                graph.add_node(to_tiles)
            if getattr(to_tiles.op, '_loose_require', False):
                loose_requires.add(to_tiles)
            objs = to_tiles.inputs or []
            for o in objs:
                if not isinstance(o, Tilesable):
                    continue
                if o not in graph:
                    graph.add_node(o)
                graph.add_edge(o, to_tiles)

                if id(o) in visited:
                    continue
                visited.add(id(o))

                q.append(o)

        visited = set()
        q = deque(graph.iter_indep())

        while q:
            node = q.popleft()
            if node in visited:
                continue
            preds = graph.predecessors(node)
            if node in loose_requires:
                accessible = any(pred in visited for pred in preds)
            else:
                accessible = all(pred in visited for pred in preds)
            if not preds or accessible:
                if node.is_coarse() and node.op:
                    tiled = self._dispatch(node.op)
                    self._assign_to([t.data for t in tiled], node.op.outputs)
                visited.add(node)
                q.extend(n for n in graph[node] if n not in visited)
            else:
                q.append(node)
                q.extend(n for n in preds if n not in visited)

        for to_tiles in loose_requires:
            tiled = self._dispatch(to_tiles.op)
            self._assign_to(tiled, to_tiles.op.outputs)

        return tiles_obj


handler = OperandTilesHandler()


def register(op, func):
    handler.register(op, func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import functools
import inspect
import json
import logging
import numbers
import os
import random
import socket
import struct
import sys
import time
import zlib
from hashlib import md5

import numpy as np

from .compat import irange, functools32, getargspec
from .utils_c import to_binary, to_str, to_text, tokenize

logger = logging.getLogger(__name__)
random.seed(int(time.time()) * os.getpid())


tokenize = tokenize


# fix encoding conversion problem under windows
if sys.platform == 'win32':  # pragma: no cover
    def _replace_default_encoding(func):
        def _fun(s, encoding=None):
            encoding = encoding or getattr(sys.stdout, 'encoding', None) or 'mbcs'
            return func(s, encoding=encoding)

        _fun.__name__ = func.__name__
        _fun.__doc__ = func.__doc__
        return _fun

    to_binary = _replace_default_encoding(to_binary)
    to_text = _replace_default_encoding(to_text)
    to_str = _replace_default_encoding(to_str)


class AttributeDict(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(
                "'AttributeDict' object has no attribute {0}".format(item))


def on_serialize_shape(shape):
    if shape:
        return tuple(s if not np.isnan(s) else -1 for s in shape)
    return shape


def on_deserialize_shape(shape):
    if shape:
        return tuple(s if s != -1 else np.nan for s in shape)
    return shape


def get_gpu_used_memory(device_id):
    import pynvml

    handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)
    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    return mem_info.used


def parse_memory_limit(value):
    if isinstance(value, numbers.Number):
        return float(value), False
    elif value.endswith('%'):
        return float(value[:-1]) / 100, True
    elif value.lower().endswith('t'):
        return float(value[:-1]) * (1024 ** 4), False
    elif value.lower().endswith('g'):
        return float(value[:-1]) * (1024 ** 3), False
    elif value.lower().endswith('m'):
        return float(value[:-1]) * (1024 ** 2), False
    elif value.lower().endswith('k'):
        return float(value[:-1]) * 1024, False
    else:
        raise ValueError('Unknown limitation value: {0}'.format(value))


def readable_size(size):
    if size < 1024:
        return size
    elif 1024 <= size < 1024 ** 2:
        return '{0:.2f}K'.format(size / 1024)
    elif 1024 ** 2 <= size < 1024 ** 3:
        return '{0:.2f}M'.format(size / (1024 ** 2))
    elif 1024 ** 3 <= size < 1024 ** 4:
        return '{0:.2f}G'.format(size / (1024 ** 3))
    else:
        return '{0:.2f}T'.format(size / (1024 ** 4))


_commit_hash, _commit_ref = None, None


def git_info():
    from ._version import get_git_info

    global _commit_hash, _commit_ref
    if _commit_ref is not None:
        if _commit_hash is None:
            return None
        return _commit_hash, _commit_ref

    git_tuple = get_git_info()
    if git_tuple is None:
        _commit_ref, _commit_hash = ':INVALID:', None
        return None
    else:
        _commit_hash, _commit_ref = git_tuple
        return git_tuple


LOW_PORT_BOUND = 10000
HIGH_PORT_BOUND = 65535
_local_occupied_ports = set()


def _get_ports_from_netstat():
    import subprocess
    p = subprocess.Popen('netstat -a -n -p tcp'.split(), stdout=subprocess.PIPE)
    p.wait()
    occupied = set()
    for line in p.stdout:
        line = to_str(line)
        if '.' not in line:
            continue
        for part in line.split():
            if '.' in part:
                _, port_str = part.rsplit('.', 1)
                if port_str == '*':
                    continue
                port = int(port_str)
                if LOW_PORT_BOUND <= port <= HIGH_PORT_BOUND:
                    occupied.add(int(port_str))
                break
    p.stdout.close()
    return occupied


def get_next_port(typ=None):
    import psutil
    try:
        conns = psutil.net_connections()
        typ = typ or socket.SOCK_STREAM
        occupied = set(sc.laddr.port for sc in conns
                       if sc.type == typ and LOW_PORT_BOUND <= sc.laddr.port <= HIGH_PORT_BOUND)
    except psutil.AccessDenied:
        occupied = _get_ports_from_netstat()

    occupied.update(_local_occupied_ports)
    randn = struct.unpack('<Q', os.urandom(8))[0]
    idx = int(randn % (1 + HIGH_PORT_BOUND - LOW_PORT_BOUND - len(occupied)))
    for i in irange(LOW_PORT_BOUND, HIGH_PORT_BOUND + 1):
        if i in occupied:
            continue
        if idx == 0:
            _local_occupied_ports.add(i)
            return i
        idx -= 1
    raise SystemError('No ports available.')


@functools32.lru_cache(200)
def mod_hash(val, modulus):
    return int(md5(to_binary(val)).hexdigest(), 16) % modulus


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


def serialize_graph(graph, compress=False):
    ser_graph = graph.to_pb().SerializeToString()
    if compress:
        ser_graph = zlib.compress(ser_graph)
    return base64.b64encode(ser_graph)


def deserialize_graph(graph_b64, graph_cls=None):
    from .serialize.protos.graph_pb2 import GraphDef
    from .graph import DirectedGraph
    graph_cls = graph_cls or DirectedGraph
    try:
        json_obj = json.loads(to_str(graph_b64))
        return graph_cls.from_json(json_obj)
    except (SyntaxError, ValueError):
        g = GraphDef()
        ser_graph = base64.b64decode(graph_b64)
        try:
            ser_graph = zlib.decompress(ser_graph)
        except zlib.error:
            pass
        g.ParseFromString(ser_graph)
        return graph_cls.from_pb(g)


def merge_tensor_chunks(input_tensor, ctx):
    from .tensor.execution.core import Executor
    from .tensor.expressions.datasource import TensorFetchChunk

    if len(input_tensor.chunks) == 1:
        return ctx[input_tensor.chunks[0].key]

    chunks = []
    for c in input_tensor.chunks:
        op = TensorFetchChunk(dtype=c.dtype, to_fetch_key=c.key)
        chunk = op.new_chunk(None, c.shape, index=c.index, _key=c.key)
        chunks.append(chunk)

    new_op = TensorFetchChunk(dtype=input_tensor.dtype, to_fetch_key=input_tensor.key)
    tensor = new_op.new_tensor(None, input_tensor.shape, chunks=chunks,
                               nsplits=input_tensor.nsplits)

    executor = Executor(storage=ctx)
    concat_result = executor.execute_tensor(tensor, concat=True)
    return concat_result[0]


if sys.version_info[0] < 3:
    def wraps(fun):
        if isinstance(fun, functools.partial):
            return lambda f: f
        return functools.wraps(fun)
else:
    wraps = functools.wraps


def calc_data_size(dt):
    if isinstance(dt, tuple):
        return sum(c.nbytes for c in dt)
    else:
        return dt.nbytes


def _get_mod_logger():
    mod_logger = None
    frame_globals = inspect.currentframe().f_back.f_globals
    for logger_name in ('logger', 'LOG', 'LOGGER'):
        if logger_name in frame_globals:
            mod_logger = frame_globals[logger_name]
            break
    return mod_logger


def log_unhandled(func):
    mod_logger = _get_mod_logger()
    if not mod_logger:
        return func

    func_name = getattr(func, '__qualname__', func.__module__ + func.__name__)
    func_args = getargspec(func)

    @wraps(func)
    def _wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:  # noqa: E722
            kwcopy = kwargs.copy()
            kwcopy.update(zip(func_args.args, args))
            if getattr(func, '__closure__', None) is not None:
                kwargs.update(zip(
                    func.__code__.co_freevars + getattr(func.__code__, 'co_cellvars', ()),
                    [getattr(c, 'cell_contents', None) for c in func.__closure__],
                ))

            messages = []
            for k, v in kwcopy.items():
                if 'key' in k:
                    messages.append('%s=%r' % (k, v))

            err_msg = 'Unexpected exception occurred in %s.' % func_name
            if messages:
                err_msg += ' ' + ' '.join(messages)
            mod_logger.exception(err_msg)
            raise
    return _wrapped



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...compat import functools32, six
from ...actors import Distributor
from ...scheduler.distributor import SchedulerDistributor
from ...worker.distributor import WorkerDistributor


def gen_distributor(scheduler_n_process, worker_n_process):
    class LocalClusterDistributor(Distributor):
        def __init__(self, n_process):
            super(LocalClusterDistributor, self).__init__(n_process)
            self._scheduler_distributor = SchedulerDistributor(scheduler_n_process)
            self._worker_distributor = WorkerDistributor(worker_n_process)

        @functools32.lru_cache(100)
        def distribute(self, uid):
            if isinstance(uid, six.string_types) and uid.startswith('w:'):
                return self._worker_distributor.distribute(uid) + scheduler_n_process

            return self._scheduler_distributor.distribute(uid)

    return LocalClusterDistributor(scheduler_n_process + worker_n_process)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ....operands import Concatenate
from ..utils import validate_axis, unify_chunks
from ..datasource import tensor as astensor
from ..core import TensorOperandMixin


class TensorConcatenate(Concatenate, TensorOperandMixin):
    def __init__(self, axis=None, dtype=None, sparse=False, **kw):
        super(TensorConcatenate, self).__init__(_axis=axis, _dtype=dtype,
                                                _sparse=sparse, **kw)

    def __call__(self, tensors):
        if len(set(t.ndim for t in tensors)) != 1:
            raise ValueError('all the input tensors must have same number of dimensions')

        axis = self._axis
        shapes = [t.shape[:axis] + t.shape[axis+1:] for t in tensors]
        if len(set(shapes)) != 1:
            raise ValueError('all the input tensor dimensions '
                             'except for the concatenation axis must match exactly')

        shape = [0 if i == axis else tensors[0].shape[i] for i in range(tensors[0].ndim)]
        shape[axis] = sum(t.shape[axis] for t in tensors)

        if any(np.isnan(s) for i, s in enumerate(shape) if i != axis):
            raise ValueError('cannot concatenate tensor with unknown shape')

        return self.new_tensor(tensors, shape=tuple(shape))

    @classmethod
    def tile(cls, op):
        from ..indexing.slice import TensorSlice

        inputs = op.inputs
        axis = op.axis

        c = itertools.count(inputs[0].ndim)
        tensor_axes = [(t, tuple(i if i != axis else next(c) for i in range(t.ndim)))
                       for t in inputs]
        inputs = unify_chunks(*tensor_axes)

        out_chunk_shape = [0 if i == axis else inputs[0].chunk_shape[i]
                           for i in range(inputs[0].ndim)]
        out_chunk_shape[axis] = sum(t.chunk_shape[axis] for t in inputs)
        out_nsplits = [None if i == axis else inputs[0].nsplits[i]
                       for i in range(inputs[0].ndim)]
        out_nsplits[axis] = tuple(itertools.chain(*[t.nsplits[axis] for t in inputs]))

        out_chunks = []
        axis_cum_chunk_shape = np.cumsum([t.chunk_shape[axis] for t in inputs])
        for out_idx in itertools.product(*[range(s) for s in out_chunk_shape]):
            axis_index = np.searchsorted(axis_cum_chunk_shape, out_idx[axis], side='right')
            t = inputs[axis_index]
            axis_inner_index = out_idx[axis] - \
                (0 if axis_index < 1 else axis_cum_chunk_shape[axis_index - 1])
            idx = out_idx[:axis] + (axis_inner_index,) + out_idx[axis+1:]
            in_chunk = t.cix[idx]
            if idx == out_idx:
                # if index is the same, just use the input chunk
                out_chunks.append(in_chunk)
            else:
                chunk_op = TensorSlice(slices=[slice(None) for _ in range(in_chunk.ndim)],
                                       dtype=in_chunk.dtype, sparse=in_chunk.op.sparse)
                out_chunk = chunk_op.new_chunk([in_chunk], in_chunk.shape, index=out_idx)

                out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  nsplits=out_nsplits, chunks=out_chunks)


def concatenate(tensors, axis=0):
    """
    Join a sequence of arrays along an existing axis.

    Parameters
    ----------
    a1, a2, ... : sequence of array_like
        The tensors must have the same shape, except in the dimension
        corresponding to `axis` (the first, by default).
    axis : int, optional
        The axis along which the tensors will be joined.  Default is 0.

    Returns
    -------
    res : Tensor
        The concatenated tensor.

    See Also
    --------
    array_split : Split a tensor into multiple sub-arrays of equal or
                  near-equal size.
    split : Split tensor into a list of multiple sub-tensors of equal size.
    hsplit : Split tensor into multiple sub-tensors horizontally (column wise)
    vsplit : Split tensor into multiple sub-tensors vertically (row wise)
    dsplit : Split tensor into multiple sub-tensors along the 3rd axis (depth).
    stack : Stack a sequence of tensors along a new axis.
    hstack : Stack tensors in sequence horizontally (column wise)
    vstack : Stack tensors in sequence vertically (row wise)
    dstack : Stack tensors in sequence depth wise (along third dimension)

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[1, 2], [3, 4]])
    >>> b = mt.array([[5, 6]])
    >>> mt.concatenate((a, b), axis=0).execute()
    array([[1, 2],
           [3, 4],
           [5, 6]])
    >>> mt.concatenate((a, b.T), axis=1).execute()
    array([[1, 2, 5],
           [3, 4, 6]])

    """
    tensors = [astensor(t) for t in tensors]

    axis = validate_axis(tensors[0].ndim, axis)
    dtype = np.result_type(*(t.dtype for t in tensors))
    sparse = all(t.issparse() for t in tensors)

    op = TensorConcatenate(axis=axis, dtype=dtype, sparse=sparse)
    return op(tensors)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .server import register_ui_handler, get_jinja_env, MarsWebAPI

_jinja_env = get_jinja_env()


def dashboard(scheduler_ip, doc):
    doc.title = 'Mars UI'

    web_api = MarsWebAPI(scheduler_ip)
    scheduler_infos = web_api.get_schedulers_info()
    worker_infos = web_api.get_workers_meta()

    scheduler_summary = {
        'count': len(scheduler_infos),
        'cpu_used': sum(si['cpu_used'] for si in scheduler_infos.values()),
        'cpu_total': sum(si['cpu_total'] for si in scheduler_infos.values()),
        'memory_used': sum(si['memory_used'] for si in scheduler_infos.values()),
        'memory_total': sum(si['memory_total'] for si in scheduler_infos.values()),
        'git_branches': set(si['git_info'] for si in scheduler_infos.values()),
    }
    worker_summary = {
        'count': len(worker_infos),
        'cpu_used': sum(wi['hardware']['cpu_used'] for wi in worker_infos.values()),
        'cpu_total': sum(wi['hardware']['cpu_total'] for wi in worker_infos.values()),
        'memory_used': sum(wi['hardware']['memory_used'] for wi in worker_infos.values()),
        'memory_total': sum(wi['hardware']['memory_total'] for wi in worker_infos.values()),
        'git_branches': set(wi['details']['git_info'] for wi in worker_infos.values()),
    }

    doc.template_variables['scheduler_summary'] = scheduler_summary
    doc.template_variables['worker_summary'] = worker_summary
    doc.template = _jinja_env.get_template('dashboard.html')


register_ui_handler('/', dashboard)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .core import Operand
from .. import opcodes as OperandDef
from ..serialize import ValueType, AnyField, KeyField, ListField, TupleField, DataTypeField, \
    StringField, Int32Field, Int64Field, BoolField


class HasInput(Operand):
    __slots__ = ()

    @property
    def input(self):
        return self._input

    def _set_inputs(self, inputs):
        super(HasInput, self)._set_inputs(inputs)
        self._input = self._inputs[0]


class Reshape(HasInput):
    _op_type_ = OperandDef.RESHAPE

    _input = KeyField('input')
    _newshape = TupleField('newshape', ValueType.int64)

    @property
    def newshape(self):
        return self._newshape


class CopyTo(Operand):
    _op_type_ = OperandDef.COPYTO

    _src = KeyField('src')
    _dst = KeyField('dest')
    _casting = StringField('casting')
    _where = KeyField('where')

    @property
    def src(self):
        return self._src

    @property
    def dst(self):
        return self._dst

    @property
    def casting(self):
        return self._casting

    @property
    def where(self):
        return self._where


class Astype(HasInput):
    _op_type_ = OperandDef.ASTYPE

    _input = KeyField('input')
    _dtype = DataTypeField('dtype')
    _casting = StringField('casting')

    @property
    def dtype(self):
        return self._dtype

    @property
    def casting(self):
        return self._casting


class Transpose(HasInput):
    _op_type_ = OperandDef.TRANSPOSE

    _input = KeyField('input')
    _axes = ListField('axes', ValueType.int32)

    @property
    def axes(self):
        return getattr(self, '_axes', None)


class Slice(HasInput):
    _op_type_ = OperandDef.SLICE

    _input = KeyField('input')
    _slices = ListField('slices')

    @property
    def slices(self):
        return self._slices


class Index(HasInput):
    _op_type_ = OperandDef.INDEX

    _input = KeyField('input')
    _indexes = ListField('indexes')

    @property
    def indexes(self):
        return self._indexes


class IndexSetValue(HasInput):
    _op_type_ = OperandDef.INDEXSETVALUE

    _input = KeyField('input')
    _indexes = ListField('indexes')
    _value = AnyField('value')

    @property
    def indexes(self):
        return self._indexes

    @property
    def value(self):
        return self._value


class UnravelIndex(HasInput):
    _op_type_ = OperandDef.UNRAVEL_INDEX

    _input = KeyField('input')
    _dims = TupleField('dims', ValueType.int32)

    @property
    def dims(self):
        return self._dims

    @property
    def output_limit(self):
        return float('inf')


class Split(HasInput):
    _op_type_ = OperandDef.ARRAY_SPLIT

    _input = KeyField('input')
    _indices_or_sections = AnyField('indices_or_sections')
    _axis = Int32Field('axis')

    @property
    def indices_or_sections(self):
        return self._indices_or_sections

    @property
    def axis(self):
        return getattr(self, '_axis', 0)

    @property
    def output_limit(self):
        return float('inf')


class BroadcastTo(HasInput):
    _op_type_ = OperandDef.BROADCAST_TO

    _input = KeyField('input')
    _shape = TupleField('shape')

    @property
    def shape(self):
        return self._shape


class Where(Operand):
    _op_type_ = OperandDef.WHERE

    _condition = KeyField('condition')
    _x = KeyField('x')
    _y = KeyField('y')

    @property
    def condition(self):
        return self._condition

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y


class Argwhere(HasInput):
    _op_type_ = OperandDef.ARGWHERE

    _input = KeyField('input')


class Squeeze(HasInput):
    _op_type_ = OperandDef.SQUEEZE

    _input = KeyField('input')
    _axis = TupleField('axis', ValueType.int32)

    @property
    def axis(self):
        return self._axis


class Digitize(HasInput):
    _op_type_ = OperandDef.DIGITIZE

    _input = KeyField('input')
    _bins = AnyField('bins')
    _right = BoolField('right')

    @property
    def bins(self):
        return self._bins

    @property
    def right(self):
        return self._right


class Repeat(HasInput):
    _op_type_ = OperandDef.REPEAT

    _input = KeyField('input')
    _repeats = AnyField('repeats')
    _axis = Int32Field('axis')

    @property
    def repeats(self):
        return self._repeats

    @property
    def axis(self):
        return self._axis


class IsIn(Operand):
    _op_type_ = OperandDef.ISIN

    _element = KeyField('element')
    _test_elements = KeyField('test_elements')
    _assume_unique = BoolField('assume_unique')
    _invert = BoolField('invert')

    @property
    def element(self):
        return self._element

    @property
    def test_elements(self):
        return self._test_elements

    @property
    def assume_unique(self):
        return self._assume_unique

    @property
    def invert(self):
        return self._invert


class SwapAxes(HasInput):
    _op_type_ = OperandDef.SWAPAXES

    _input = KeyField('input')
    _axis1 = Int32Field('axis1')
    _axis2 = Int32Field('axis2')

    @property
    def axis1(self):
        return self._axis1

    @property
    def axis2(self):
        return self._axis2


class Stack(Operand):
    _op_type_ = OperandDef.STACK

    _axis = Int32Field('axis')

    @property
    def axis(self):
        return self._axis


class Rechunk(HasInput):
    _op_type_ = OperandDef.RECHUNK

    _input = KeyField('input')
    _chunk_size = AnyField('chunk_size')
    _threshold = Int32Field('threshold')
    _chunk_size_limit = Int64Field('chunk_size_limit')

    @property
    def chunk_size(self):
        return self._chunk_size

    @property
    def threshold(self):
        return self._threshold

    @property
    def chunk_size_limit(self):
        return self._chunk_size_limit



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import operator
import heapq
import itertools

import numpy as np

from ....config import options
from ....compat import six, izip, lzip, reduce
from ....operands import Rechunk
from ..utils import decide_chunk_sizes, calc_sliced_size
from ..core import TensorOperandMixin


class TensorRechunk(Rechunk, TensorOperandMixin):
    def __init__(self, chunk_size=None, threshold=None, chunk_size_limit=None, dtype=None, **kw):
        super(TensorRechunk, self).__init__(_chunk_size=chunk_size, _threshold=threshold,
                                            _chunk_size_limit=chunk_size_limit, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorRechunk, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, tensor):
        return self.new_tensor([tensor], tensor.shape)

    @classmethod
    def tile(cls, op):
        new_chunk_size = op.chunk_size
        steps = plan_rechunks(op.inputs[0], new_chunk_size,
                              threshold=op.threshold,
                              chunk_size_limit=op.chunk_size_limit)
        tensor = op.outputs[0]
        for c in steps:
            tensor = compute_rechunk(tensor.inputs[0], c)

        return [tensor]


def rechunk(tensor, chunk_size, threshold=None, chunk_size_limit=None):
    chunk_size = _get_nsplits(tensor, chunk_size)
    if chunk_size == tensor.nsplits:
        return tensor

    op = TensorRechunk(chunk_size, threshold, chunk_size_limit, dtype=tensor.dtype)
    return op(tensor)


# -----------------------------------------------------------------------------------
# We adopt the iterative rechunk logic from dask, see: https://github.com/dask/dask
# -----------------------------------------------------------------------------------


def _get_nsplits(tensor, new_chunk_size):
    if isinstance(new_chunk_size, dict):
        chunk_size = list(tensor.nsplits)
        for idx, c in six.iteritems(new_chunk_size):
            chunk_size[idx] = c
    else:
        chunk_size = new_chunk_size

    return decide_chunk_sizes(tensor.shape, chunk_size, tensor.dtype.itemsize)


def _largest_chunk_size(nsplits):
    return reduce(operator.mul, map(max, nsplits))


def _chunk_number(nsplits):
    return reduce(operator.mul, map(len, nsplits))


def _estimate_graph_size(old_chunk_size, new_chunk_size):
    return reduce(operator.mul,
                  (len(oc) + len(nc) for oc, nc in zip(old_chunk_size, new_chunk_size)))


def plan_rechunks(tensor, new_chunk_size, threshold=None, chunk_size_limit=None):
    threshold = threshold or options.tensor.rechunk.threshold
    chunk_size_limit = chunk_size_limit or options.tensor.rechunk.chunk_size_limit

    if len(new_chunk_size) != tensor.ndim:
        raise ValueError('Provided chunks should have %d dimensions, got %d' % (
            tensor.ndim, len(new_chunk_size)))

    steps = []

    chunk_size_limit /= tensor.dtype.itemsize
    chunk_size_limit = max([int(chunk_size_limit),
                            _largest_chunk_size(tensor.nsplits),
                            _largest_chunk_size(new_chunk_size)])

    graph_size_threshold = threshold * (_chunk_number(tensor.nsplits) + _chunk_number(new_chunk_size))

    chunk_size = curr_chunk_size = tensor.nsplits
    first_run = True
    while True:
        graph_size = _estimate_graph_size(chunk_size, new_chunk_size)
        if graph_size < graph_size_threshold:
            break
        if not first_run:
            chunk_size = _find_split_rechunk(curr_chunk_size, new_chunk_size, graph_size * threshold)
        chunks_size, memory_limit_hit = _find_merge_rechunk(chunk_size, new_chunk_size, chunk_size_limit)
        if chunk_size == curr_chunk_size or chunk_size == new_chunk_size:
            break
        steps.append(chunk_size)
        curr_chunk_size = chunk_size
        if not memory_limit_hit:
            break
        first_run = False

    return steps + [new_chunk_size]


def _find_split_rechunk(old_chunk_size, new_chunk_size, graph_size_limit):
    """
        Find an intermediate rechunk that would merge some adjacent blocks
        together in order to get us nearer the *new_chunk_size* target, without
        violating the *graph_size_limit* (in number of elements).
        """
    ndim = len(old_chunk_size)

    old_largest_width = [max(c) for c in old_chunk_size]
    new_largest_width = [max(c) for c in new_chunk_size]

    graph_size_effect = {
        dim: len(nc) / len(oc)
        for dim, (oc, nc) in enumerate(zip(old_chunk_size, new_chunk_size))
    }

    block_size_effect = {
        dim: new_largest_width[dim] / old_largest_width[dim]
        for dim in range(ndim)
    }

    # Our goal is to reduce the number of nodes in the rechunk graph
    # by merging some adjacent chunks, so consider dimensions where we can
    # reduce the # of chunks
    merge_candidates = [dim for dim in range(ndim)
                        if graph_size_effect[dim] <= 1.0]

    # Merging along each dimension reduces the graph size by a certain factor
    # and increases memory largest block size by a certain factor.
    # We want to optimize the graph size while staying below the given
    # graph_size_limit.  This is in effect a knapsack problem, except with
    # multiplicative values and weights.  Just use a greedy algorithm
    # by trying dimensions in decreasing value / weight order.
    def key(k):
        gse = graph_size_effect[k]
        bse = block_size_effect[k]
        if bse == 1:
            bse = 1 + 1e-9
        return np.log(gse) / np.log(bse)

    sorted_candidates = sorted(merge_candidates, key=key)

    largest_block_size = reduce(operator.mul, old_largest_width)

    chunk_size = list(old_chunk_size)
    memory_limit_hit = False

    for dim in sorted_candidates:
        # Examine this dimension for possible graph reduction
        new_largest_block_size = (
            largest_block_size * new_largest_width[dim] // old_largest_width[dim])
        if new_largest_block_size <= graph_size_limit:
            # Full replacement by new chunks is possible
            chunk_size[dim] = new_chunk_size[dim]
            largest_block_size = new_largest_block_size
        else:
            # Try a partial rechunk, dividing the new chunks into
            # smaller pieces
            largest_width = old_largest_width[dim]
            chunk_limit = int(graph_size_limit * largest_width / largest_block_size)
            c = _divide_to_width(new_chunk_size[dim], chunk_limit)
            if len(c) <= len(old_chunk_size[dim]):
                # We manage to reduce the number of blocks, so do it
                chunk_size[dim] = c
                largest_block_size = largest_block_size * max(c) // largest_width

            memory_limit_hit = True

    assert largest_block_size == _largest_chunk_size(chunk_size)
    assert largest_block_size <= graph_size_limit
    return tuple(chunk_size), memory_limit_hit


def _merge_to_number(desired_chunk_size, max_number):
    """ Minimally merge the given chunks so as to drop the number of
        chunks below *max_number*, while minimizing the largest width.
        """
    if len(desired_chunk_size) <= max_number:
        return desired_chunk_size

    distinct = set(desired_chunk_size)
    if len(distinct) == 1:
        # Fast path for homogeneous target, also ensuring a regular result
        w = distinct.pop()
        n = len(desired_chunk_size)
        total = n * w

        desired_width = total // max_number
        width = w * (desired_width // w)
        adjust = (total - max_number * width) // w

        return (width + w,) * adjust + (width,) * (max_number - adjust)

    nmerges = len(desired_chunk_size) - max_number

    heap = [(desired_chunk_size[i] + desired_chunk_size[i + 1], i, i + 1)
            for i in range(len(desired_chunk_size) - 1)]
    heapq.heapify(heap)

    chunk_size = list(desired_chunk_size)

    while nmerges > 0:
        # Find smallest interval to merge
        width, i, j = heapq.heappop(heap)
        # If interval was made invalid by another merge, recompute
        # it, re-insert it and retry.
        if chunk_size[j] == 0:
            j += 1
            while chunk_size[j] == 0:
                j += 1
            heapq.heappush(heap, (chunk_size[i] + chunk_size[j], i, j))
            continue
        elif chunk_size[i] + chunk_size[j] != width:
            heapq.heappush(heap, (chunk_size[i] + chunk_size[j], i, j))
            continue
        # Merge
        assert chunk_size[i] != 0
        chunk_size[i] = 0  # mark deleted
        chunk_size[j] = width
        nmerges -= 1

    return tuple(filter(None, chunk_size))


def _divide_to_width(desired_chunk_size, max_width):
    """ Minimally divide the given chunks so as to make the largest chunk
    width less or equal than *max_width*.
    """
    chunk_size = []
    for c in desired_chunk_size:
        nb_divides = int(np.ceil(c / max_width))
        for i in range(nb_divides):
            n = c // (nb_divides - i)
            chunk_size.append(n)
            c -= n
        assert c == 0
    return tuple(chunk_size)


def _find_merge_rechunk(old_chunk_size, new_chunk_size, chunk_size_limit):
    """
        Find an intermediate rechunk that would merge some adjacent blocks
        together in order to get us nearer the *new_chunks* target, without
        violating the *chunk_size_limit* (in number of elements).
        """
    ndim = len(old_chunk_size)

    old_largest_width = [max(c) for c in old_chunk_size]
    new_largest_width = [max(c) for c in new_chunk_size]

    graph_size_effect = {
        dim: len(nc) / len(oc)
        for dim, (oc, nc) in enumerate(zip(old_chunk_size, new_chunk_size))
    }

    block_size_effect = {
        dim: new_largest_width[dim] / old_largest_width[dim]
        for dim in range(ndim)
    }

    # Our goal is to reduce the number of nodes in the rechunk graph
    # by merging some adjacent chunks, so consider dimensions where we can
    # reduce the # of chunks
    merge_candidates = [dim for dim in range(ndim)
                        if graph_size_effect[dim] <= 1.0]

    # Merging along each dimension reduces the graph size by a certain factor
    # and increases memory largest block size by a certain factor.
    # We want to optimize the graph size while staying below the given
    # chunk_size_limit.  This is in effect a knapsack problem, except with
    # multiplicative values and weights.  Just use a greedy algorithm
    # by trying dimensions in decreasing value / weight order.
    def key(k):
        gse = graph_size_effect[k]
        bse = block_size_effect[k]
        if bse == 1:
            bse = 1 + 1e-9
        return np.log(gse) / np.log(bse)

    sorted_candidates = sorted(merge_candidates, key=key)

    largest_block_size = reduce(operator.mul, old_largest_width)

    chunk_size = list(old_chunk_size)
    memory_limit_hit = False

    for dim in sorted_candidates:
        # Examine this dimension for possible graph reduction
        new_largest_block_size = (
            largest_block_size * new_largest_width[dim] // old_largest_width[dim])
        if new_largest_block_size <= chunk_size_limit:
            # Full replacement by new chunks is possible
            chunk_size[dim] = new_chunk_size[dim]
            largest_block_size = new_largest_block_size
        else:
            # Try a partial rechunk, dividing the new chunks into
            # smaller pieces
            largest_width = old_largest_width[dim]
            chunk_limit = int(chunk_size_limit * largest_width / largest_block_size)
            c = _divide_to_width(new_chunk_size[dim], chunk_limit)
            if len(c) <= len(old_chunk_size[dim]):
                # We manage to reduce the number of blocks, so do it
                chunk_size[dim] = c
                largest_block_size = largest_block_size * max(c) // largest_width

            memory_limit_hit = True

    assert largest_block_size == _largest_chunk_size(chunk_size)
    assert largest_block_size <= chunk_size_limit
    return tuple(chunk_size), memory_limit_hit


def compute_rechunk(tensor, chunk_size):
    from ..indexing.slice import TensorSlice
    from ..merge.concatenate import TensorConcatenate

    nsplits = tensor.nsplits
    truncated = [[0, None] for _ in range(tensor.ndim)]
    result_slices = []
    for dim, old_chunk_size, new_chunk_size in izip(itertools.count(0), nsplits, chunk_size):
        slices = []
        for rest in new_chunk_size:
            dim_slices = []
            while rest > 0:
                old_idx, old_start = truncated[dim]
                old_size = old_chunk_size[old_idx] - (old_start or 0)
                if old_size < rest:
                    dim_slices.append((old_idx, slice(old_start, None, None)))
                    rest -= old_size
                    truncated[dim] = old_idx + 1, None
                else:
                    end = rest if old_start is None else rest + old_start
                    if end >= old_chunk_size[old_idx]:
                        truncated[dim] = old_idx + 1, None
                    else:
                        truncated[dim] = old_idx, end
                    end = None if end == old_chunk_size[old_idx] else end
                    dim_slices.append((old_idx, slice(old_start, end)))
                    rest -= old_size
            slices.append(dim_slices)
        result_slices.append(slices)

    result_chunks = []
    idxes = itertools.product(*[range(len(c)) for c in chunk_size])
    chunk_slices = itertools.product(*result_slices)
    chunk_shapes = itertools.product(*chunk_size)
    for idx, chunk_slice, chunk_shape in izip(idxes, chunk_slices, chunk_shapes):
        to_merge = []
        merge_idxes = itertools.product(*[range(len(i)) for i in chunk_slice])
        for merge_idx, index_slices in izip(merge_idxes, itertools.product(*chunk_slice)):
            chunk_index, chunk_slice = lzip(*index_slices)
            old_chunk = tensor.cix[chunk_index]
            merge_chunk_shape = tuple(calc_sliced_size(s, chunk_slice[0]) for s in old_chunk.shape)
            merge_chunk_op = TensorSlice(chunk_slice, dtype=old_chunk.dtype)
            merge_chunk = merge_chunk_op.new_chunk([old_chunk], merge_chunk_shape, index=merge_idx)
            to_merge.append(merge_chunk)
        if len(to_merge) == 1:
            chunk_op = to_merge[0].op.copy()
            out_chunk = chunk_op.new_chunk(to_merge[0].op.inputs, chunk_shape, index=idx)
            result_chunks.append(out_chunk)
        else:
            chunk_op = TensorConcatenate(dtype=to_merge[0].dtype)
            out_chunk = chunk_op.new_chunk(to_merge, chunk_shape, index=idx)
            result_chunks.append(out_chunk)

    op = TensorRechunk(chunk_size)
    return op.new_tensor([tensor], tensor.shape, dtype=tensor.dtype,
                         nsplits=chunk_size, chunks=result_chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils import validate_axis
from ..datasource import tensor as astensor


def diff(a, n=1, axis=-1):
    """
    Calculate the n-th discrete difference along the given axis.

    The first difference is given by ``out[n] = a[n+1] - a[n]`` along
    the given axis, higher differences are calculated by using `diff`
    recursively.

    Parameters
    ----------
    a : array_like
        Input tensor
    n : int, optional
        The number of times values are differenced. If zero, the input
        is returned as-is.
    axis : int, optional
        The axis along which the difference is taken, default is the
        last axis.

    Returns
    -------
    diff : Tensor
        The n-th differences. The shape of the output is the same as `a`
        except along `axis` where the dimension is smaller by `n`. The
        type of the output is the same as the type of the difference
        between any two elements of `a`. This is the same as the type of
        `a` in most cases. A notable exception is `datetime64`, which
        results in a `timedelta64` output tensor.

    See Also
    --------
    gradient, ediff1d, cumsum

    Notes
    -----
    Type is preserved for boolean tensors, so the result will contain
    `False` when consecutive elements are the same and `True` when they
    differ.

    For unsigned integer tensors, the results will also be unsigned. This
    should not be surprising, as the result is consistent with
    calculating the difference directly:

    >>> import mars.tensor as mt

    >>> u8_arr = mt.array([1, 0], dtype=mt.uint8)
    >>> mt.diff(u8_arr).execute()
    array([255], dtype=uint8)
    >>> (u8_arr[1,...] - u8_arr[0,...]).execute()
    255

    If this is not desirable, then the array should be cast to a larger
    integer type first:

    >>> i16_arr = u8_arr.astype(mt.int16)
    >>> mt.diff(i16_arr).execute()
    array([-1], dtype=int16)

    Examples
    --------
    >>> x = mt.array([1, 2, 4, 7, 0])
    >>> mt.diff(x).execute()
    array([ 1,  2,  3, -7])
    >>> mt.diff(x, n=2).execute()
    array([  1,   1, -10])

    >>> x = mt.array([[1, 3, 6, 10], [0, 5, 6, 8]])
    >>> mt.diff(x).execute()
    array([[2, 3, 4],
           [5, 1, 2]])
    >>> mt.diff(x, axis=0).execute()
    array([[-1,  2,  0, -2]])

    >>> x = mt.arange('1066-10-13', '1066-10-16', dtype=mt.datetime64)
    >>> mt.diff(x).execute()
    array([1, 1], dtype='timedelta64[D]')

    """
    a = astensor(a)
    n = int(n)

    axis = validate_axis(a.ndim, axis)
    slc1 = (slice(None),) * axis + (slice(1, None),)
    slc2 = (slice(None),) * axis + (slice(-1),)

    for _ in range(n):
        a = a[slc1] - a[slc2]

    return a



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from .array import SparseNDArray
from .. import six


class COONDArray(SparseNDArray):
    __slots__ = ('indices', 'data', 'shape', 'dtype')
    """
    SparseNDArray in COO format.
    Stores indices in a 2-D array, indices, of size n * ndims, where n is the number of
    non-zero values and ndims is the number of dimensions of
    the SparseNDArray.
    For example,
    indices = [(1,3,4), (2,1,3), (3, 2, 2)],
    data = [11, 12, 13],
    shape = [4, 5, 6],
    together they represent a sparse array of size 4*5*6
    with data 11, 12, 13 at coordinate (1,3,4), (2,1,3) and (3, 2, 2).

    Attributes:
    -----------
    data: ndarray
        A 1-D array of size n, where n is the number of elements
    indices: ndarray
        A 2-D array of size n * ndims, where n is the number of non-zero values
            and ndims is the number of dimensions of the SparseNDArray.
    shape: ndarray
        A 1-D array of size ndims, where ndims is the number of dimensions
            of the sparse array in the dense form.
    """

    def __init__(self, indices=(), data=(), shape=(), dtype=np.int64):
        try:
            self.dtype = data.dtype
        except AttributeError:
            self.dtype = dtype
        self.indices = np.asarray(indices)
        self.data = np.asarray(data, dtype=self.dtype)
        self.shape = np.asarray(shape)

    # return raw components of a COONDArray instance
    @property
    def raw(self):
        return self.indices, self.data, self.shape

    @property
    def ndim(self):
        return self.indices.shape[1]

    @property
    def nnz(self):
        return self.indices.shape[0]

    # return a copy of a COONDArray instance
    def copy(self):
        return COONDArray(self.indices, self.data, self.shape)

    # Return a copy of nd.ndarray of self
    def toarray(self):
        # allocate dense_arr
        dense_arr = np.zeros(shape=self.shape, dtype=self.data.dtype, order='C')

        # Assign values to corresponding coordinates in dense array
        # Each column vector contains coordinates of one and only one dim.
        if len(self.data) == 0 and len(self.indices) == 0:
            return dense_arr
        dense_arr[tuple(self.indices.T)] = self.data
        # ( Written in plainer words:
        # dense_arr[
        #     <split n-dim coords into n column vectors>,
        #     number of dims, split along axis 1
        # ] = transform_into_column_vectors(self.values)
        # P.S. Each column vector contains coordinates of a dim. )
        return dense_arr

    # Per scipy sparse issparse / scipy matrix issparse
    def issparse(self):
        if isinstance(self, COONDArray):
            return True

    def mask_matches(self, other):
        mask = (self.indices[:, np.newaxis, :]
                == other.indices[np.newaxis, :, :]).all(axis=2)
        matches = np.where(mask == True)
        # Check if the entire row is false.
        # If so, then the coordinate at the corresponding position
        # in the self.indices does not have a matching coordinate in other.indices
        self_non_matches = np.where(~mask.any(axis=1))
        # Check if the entire column is false;.
        # If so, then the coordinate at corresponding position
        # in the other.indices does not have a matching coordinate in self.indices
        other_non_matches = np.where(~mask.any(axis=0))
        matches = np.asarray(matches)
        return matches, self_non_matches, other_non_matches

    def __add__(self, other, increment_all=False):
        """
        Perform elementwise addition of two sparse tensor.
        ind stands for index, inds stands for indices, dct stands for dictionary.

        :param other:
        :return: the sum of two SparseNDArray as an object.
        OUTPUT TYPE:
        Sparse + Sparse -> Sparse
        Sparse + Dense -> Dense
        Sparse + Constant -> Dense (following conventions of other libraries)
        """
        if isinstance(other, SparseNDArray):
            matches, self_non_matches, other_non_matches = self.mask_matches(other)

            # calculate sums of values at matching coordinates
            matching_vals = self.data[matches[0]] + other.data[matches[1]]
            # document matching coordinates
            matching_indices = self.indices[matches[0]]

            # append non-matching coordinates in self to existing indices
            new_indices = np.append(matching_indices,
                                    self.indices[self_non_matches], axis=0)
            # append non-matching coordinates in other to exisiting indices
            new_indices = np.append(new_indices,
                                    other.indices[other_non_matches], axis=0)
            # append values at non-matching coordinates in self to existing values
            new_vals = np.append(matching_vals,
                                 self.data[self_non_matches], axis=0)
            # append values at non-matching coordinates in other to existing values
            new_vals = np.append(new_vals,
                                 other.data[other_non_matches], axis=0)

            return COONDArray(new_indices, new_vals, self.shape)

        elif isinstance(other, np.ndarray):
            result = np.copy(other)
            for i in range(len(self.indices)):
                result[tuple(self.indices[i])] += self.data[i]
            return result

        elif isinstance(other, (int, float, np.int, np.float)):
            result = np.ones(shape=self.shape) * other
            for i in range(len(self.indices)):
                result[self.indices[i]] += self.data[i]
            return result

        else:
            return NotImplemented
            # raise TypeError(
            #     "The added is not (dense)np.ndarray, "
            #     "(sparse)COONDArray or (constant)int/float. ")

    def __sub__(self, other, increment_all=False):
        """
        Perform element-wise subtraction of two sparse tensor.
        （numpy function to use: setdiff1d, in1d, isin.）

        If having difficulty of understanding and/or doubts about any
        implementation decisions in this function, refer to the
        comment block in __add__.

        :param other: what to subtract
        :return: the sum of two SparseNDArray as an object.
                    Sparse - Sparse -> Sparse
                    Sparse - Dense -> Sparse
                    Sparse - Constant -> Dense

        Refer to additional comments on decision-makings of implementation
        at the corresponding location in __add__ function.
        """
        if isinstance(other, SparseNDArray):
            matches, self_non_matches, other_non_matches = self.mask_matches(other)

            # calculate sums of values at matching coordinates
            matching_vals = self.data[matches[0]] - other.data[matches[1]]
            # document matching coordinates
            matching_indices = self.indices[matches[0]]

            # append non-matching coordinates in self to existing indices
            new_indices = np.append(matching_indices,
                                    self.indices[self_non_matches], axis=0)
            # append non-matching coordinates in other to exisiting indices
            new_indices = np.append(new_indices,
                                    other.indices[other_non_matches], axis=0)
            # append values at non-matching coordinates in self to existing values
            new_vals = np.append(matching_vals,
                                 self.data[self_non_matches], axis=0)
            # append values at non-matching coordinates in other to existing values
            new_vals = np.append(new_vals,
                                 (-1) * other.data[other_non_matches], axis=0)

            return COONDArray(new_indices, new_vals, self.shape)

        elif isinstance(other, np.ndarray):
            result = np.copy(other)
            for i in range(len(self.indices)):
                result[tuple(self.indices[i])] -= self.data[i]
            return np.negative(result)

        elif isinstance(other, (int, float, np.int, np.float)):
            result = np.ones(shape=self.shape) * other
            for i in range(len(self.indices)):
                result[self.indices[i]] -= self.data[i]
            return np.negative(result)

        else:
            return NotImplemented
            # Equivalent to:
            # raise TypeError(
            #     "The subtracted is not (dense)np.ndarray, "
            #     "(sparse)COONDArray or (constant)int/float. ")

    def __mul__(self, other):
        """
        Element-wise multiplication between two matrices, self and other.
        Self(self) is always assumed to be sparse.
        Other(other) can be one of the three types: sparse COONDArray,
            dense np.ndarray or any other numerical values.

        :param other: the multiplied
        :return: element-wise product of self and other
                    (S: Sparse, D: Dense, C: Constant)
                    Sparse * Sparse -> Sparse
                    Sparse * Dense -> Sparse
                    Sparse * Constant -> Sparse

        More on my __mul__ implementation:
            When operating on operands in the COO format, one must pay attention to matches of coordinates.
            For different operators, logic on manipulating data from left and right side of operators may be
            subtly different, yet these differences are significant.
            For summations and subtractions, they are analogous to LOGIC OR, since we book-keep both sides on
            either '+' or '-' whether a coordinate
            has a match or not.
            For multiplications, it is analogous to LOGIC AND, since we book-keep only matching coordinates
            and their values; the rest of them are made irrelevant by being multiplied by zero.
            Division is a bit complicated. It shares some similarities with multiplications, but, when other
            contains 0, ZeroDivisionError will be raised and values of those elements will be np.inf.

        """
        # if isinstance(other, SparseNDArray) and self.shape == other.shape:
        if isinstance(other, SparseNDArray) and (self.shape == other.shape).all():
            mask = (self.indices[:, np.newaxis, :] == other.indices[np.newaxis, :, :]).all(axis=2)
            matches = np.where(mask == True)
            matches = np.asarray(matches)

            matching_vals = self.data[matches[0]] * other.data[matches[1]]
            matching_indices = self.indices[matches[0]]

            new_vals = matching_vals
            new_indices = matching_indices

            return COONDArray(new_indices, new_vals, self.shape)

        elif isinstance(other, np.ndarray):
            data = np.copy(self.data)
            for i in range(len(data)):
                data[i] *= other[tuple(self.indices[i])]
            return COONDArray(self.indices, data, self.shape)

        elif isinstance(other, (int, float, np.int, np.float)):
            if other == 0 or other == float(0):
                return COONDArray(indices=[], data=[], shape=self.shape)
            elif other == 1 or other == float(1):
                return COONDArray(self.indices, self.data, self.shape)
            else:
                data = list(map(lambda v: v * other, self.data))
                return COONDArray(self.indices, data, self.shape)

        else:
            return NotImplemented
            # Equivalent to:
            # raise TypeError("The multiplying is not (dense)np.ndarray, (sparse)COONDArray or (constant)int/float.")

    def __truediv__(self, other):
        """
        Element-wise division between two matrices, self and other.
        Self(self) is always assumed to be sparse.
        Other(other) can be one of these types: sparse COONDArray,
            dense np.ndarray or any other numerical values.

        :param other: the divisor
        :return: element-wise product of self and other
                    S: Sparse, D: Dense, C: Constant
                    Sparse / Sparse -> Sparse
                    Sparse / Dense -> Sparse
                    Sparse * Constant -> Sparse
        """

        if isinstance(other, COONDArray):
            if six.PY2:
                return np.true_divide(self.toarray(), other.toarray())
            else:
                return self.toarray() / other.toarray()

        elif isinstance(other, np.ndarray):
            if six.PY2:
                return np.true_divide(self.toarray(), other)
            else:
                return self.toarray() / other

        elif isinstance(other, (int, float, np.int, np.float)):
            if other == 0 or other == 0.0:
                return NotImplemented
                # Equivalent to:
                # raise ValueError("A divisor should not be zero")

            elif other == 1 or other == 1.0:
                return COONDArray(self.indices,
                                  np.asarray(self.data, dtype=type(other)), self.shape)

            else:
                if six.PY2:
                    data = np.true_divide(self.data, float(other))
                else:
                    data = np.asarray(self.data) / float(other)
                return COONDArray(self.indices, data, self.shape)

        else:
            return NotImplemented
            # Equivalent to:
            # raise TypeError("The divisor is not (dense)np.ndarray, (sparse)COONDArray or (constant)int/float. ")

    # Support function for python 2.7
    # IN THE FUTURE: __div__ returns the result of the function call to __floordiv__
    def __div__(self, other):
        return self.__truediv__(other)

    def transpose(self, axes=None):
        """
        transpose a COONDArray instance.
        (Exchange values of coordinates given axes (index of coordinate values))

        :param axes: axes whose coordinates are changed.
        :return: not returning anyting. Instead, states of the calling objecting
                    is updated. Change indices of the calling object.
        """
        # CASE: incorrect input type

        if axes is None:
            # indices = list(map(lambda c: tuple(c[-1::-1]), self.indices))
            return COONDArray(self.indices[:, -1::-1], self.data, self.shape[-1::-1])

        elif not isinstance(axes, tuple):
            # return NotImplemented
            # Equivalent to:
            raise TypeError("axes should be either None or a tuple of length 2. ")

        elif isinstance(axes, tuple) and len(axes) != 2:
            # return NotImplemented
            # Equivalent to:
            raise ValueError("If a tuple is given, its length should be 2. ")

        elif isinstance(axes, tuple) and len(axes) == 2:
            indices = np.copy(self.indices)
            shape = np.copy(self.shape)

            indices[:, [axes[0], axes[1]]] = indices[:, [axes[1], axes[0]]]
            shape[[axes[0], axes[1]]] = shape[[axes[1], axes[0]]]

            return COONDArray(indices, self.data, shape)

        else:
            # return NotImplemented
            raise TypeError("Some unknown type is given as input! ")


# Per SciPy csr_matrix issparse().
def is_coo(x):
    if isinstance(x, COONDArray):
        return True



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import keyword
import inspect
from collections import namedtuple

from . import six


if six.PY3:
    def isvalidattr(ident):
        return not keyword.iskeyword(ident) and ident.isidentifier()
else:
    def isvalidattr(ident):
        """Determines, if string is valid Python identifier."""

        # Smoke test - if it's not string, then it's not identifier, but we don't
        # want to just silence exception. It's better to fail fast.
        if not isinstance(ident, str):
            raise TypeError('expected str, but got {!r}'.format(type(ident)))

        # Resulting AST of simple identifier is <Module [<Expr <Name "foo">>]>
        try:
            root = ast.parse(ident)
        except SyntaxError:
            return False

        if not isinstance(root, ast.Module) \
                or len(root.body) != 1 \
                or not isinstance(root.body[0], ast.Expr) \
                or not isinstance(root.body[0].value, ast.Name) \
                or root.body[0].value.id != ident:
            return False

        return True


if six.PY3:
    def dir2(obj):
        return object.__dir__(obj)
else:
    # http://www.quora.com/How-dir-is-implemented-Is-there-any-PEP-related-to-that
    def get_attrs(obj):
        import types
        if not hasattr(obj, '__dict__'):
            return []  # slots only
        if not isinstance(obj.__dict__, (dict, types.DictProxyType)):
            raise TypeError("%s.__dict__ is not a dictionary"
                            "" % obj.__name__)
        return obj.__dict__.keys()

    def dir2(obj):
        attrs = set()
        if not hasattr(obj, '__bases__'):
            # obj is an instance
            if not hasattr(obj, '__class__'):
                # slots
                return sorted(get_attrs(obj))
            klass = obj.__class__
            attrs.update(get_attrs(klass))
        else:
            # obj is a class
            klass = obj

        for cls in klass.__bases__:
            attrs.update(get_attrs(cls))
            attrs.update(dir2(cls))
        attrs.update(get_attrs(obj))
        return sorted(list(attrs))


def _getargspec(func):
    ArgSpec = namedtuple('ArgSpec', 'args varargs keywords defaults')

    args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = \
        getfullargspec(func)
    if kwonlyargs or ann:
        raise ValueError("Function has keyword-only arguments or annotations"
                         ", use getfullargspec() API which can support them")
    return ArgSpec(args, varargs, varkw, defaults)


def _getfullargspec(func):  # noqa: C901
    from inspect import Parameter
    FullArgSpec = namedtuple('FullArgSpec',
                             'args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations')

    sig = inspect.signature(func)
    args = []
    varargs = None
    varkw = None
    kwonlyargs = []
    annotations = {}
    defaults = ()
    kwdefaults = {}

    if sig.return_annotation is not sig.empty:
        annotations['return'] = sig.return_annotation

    for param in sig.parameters.values():
        kind = param.kind
        name = param.name

        if kind is Parameter.POSITIONAL_ONLY:
            args.append(name)
        elif kind is Parameter.POSITIONAL_OR_KEYWORD:
            args.append(name)
            if param.default is not param.empty:
                defaults += (param.default,)
        elif kind is Parameter.VAR_POSITIONAL:
            varargs = name
        elif kind is Parameter.KEYWORD_ONLY:
            kwonlyargs.append(name)
            if param.default is not param.empty:
                kwdefaults[name] = param.default
        elif kind is Parameter.VAR_KEYWORD:
            varkw = name

        if param.annotation is not param.empty:
            annotations[name] = param.annotation

    if not kwdefaults:
        # compatibility with 'func.__kwdefaults__'
        kwdefaults = None

    if not defaults:
        # compatibility with 'func.__defaults__'
        defaults = None

    return FullArgSpec(args, varargs, varkw, defaults,
                       kwonlyargs, kwdefaults, annotations)


if hasattr(inspect, 'signature'):
    getargspec, getfullargspec = _getargspec, _getfullargspec
else:
    from inspect import getargspec  # noqa: F401
    if hasattr(inspect, 'getfullargspec'):
        from inspect import getfullargspec
    else:
        getfullargspec = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ....operands import Reshape
from ....compat import six, izip
from ..datasource import tensor as astensor
from ..core import TensorOperandMixin


class TensorReshape(Reshape, TensorOperandMixin):
    def __init__(self, newshape=None, dtype=None, **kw):
        super(TensorReshape, self).__init__(_newshape=newshape, _dtype=dtype, **kw)

    def _set_inputs(self, inputs):
        super(TensorReshape, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        return self.new_tensor([a], self._newshape)

    @staticmethod
    def _gen_reshape_rechunk_nsplits(old_shape, new_shape, nsplits):
        old_idx = len(old_shape) - 1
        new_idx = len(new_shape) - 1
        rechunk_nsplists = [None for _ in old_shape]
        reshape_nsplists = [None for _ in new_shape]

        while old_idx >= 0 or new_idx >= 0:
            old_dim_size = old_shape[old_idx]
            new_dim_size = new_shape[new_idx]

            if old_dim_size == new_dim_size:
                # nothing need to do
                rechunk_nsplists[old_idx] = nsplits[old_idx]
                reshape_nsplists[new_idx] = nsplits[old_idx]
                old_idx -= 1
                new_idx -= 1
                continue

            if old_dim_size == 1:
                rechunk_nsplists[old_idx] = (1,)
                old_idx -= 1
            elif new_dim_size == 1:
                reshape_nsplists[new_idx] = (1,)
                new_idx -= 1
            elif old_dim_size < new_dim_size:
                left_old_idx = old_idx - 1
                while left_old_idx >= 0 and \
                        np.prod(old_shape[left_old_idx: old_idx + 1]) < new_dim_size:
                    left_old_idx -= 1
                if np.prod(old_shape[left_old_idx: old_idx + 1]) != new_dim_size:
                    raise ValueError('shapes not compatible')

                for i in range(left_old_idx + 1, old_idx + 1):
                    # rechunk the higher dimension into 1 chunk
                    # e.g. ((2, 2, 2), [(3, 3), (4, 4))] -> [6, 8]
                    rechunk_nsplists[i] = (old_shape[i],)

                chunk_reduce = np.prod([len(c) for c in nsplits[left_old_idx + 1: old_idx + 1]])
                # cause the higher dimension has been concatenated,
                # the lowest dimension should be expanded to reduce size
                rechunk_nsplists[left_old_idx] = \
                    TensorReshape._expand_nsplit_by_reduce(nsplits[left_old_idx], chunk_reduce)

                size_reduce = np.prod(old_shape[left_old_idx + 1: old_idx + 1])
                reshape_nsplists[new_idx] = tuple(size_reduce * c for c in rechunk_nsplists[left_old_idx])

                old_idx = left_old_idx - 1
                new_idx -= 1
            else:
                assert old_dim_size > new_dim_size
                lef_new_idx = new_idx - 1
                while lef_new_idx >= 0 and \
                        np.prod(new_shape[lef_new_idx: new_idx +1]) < old_dim_size:
                    lef_new_idx -= 1
                if np.prod(new_shape[lef_new_idx: new_idx + 1]) != old_dim_size:
                    raise ValueError('shapes not compatible')

                chunk_expand = np.prod(new_shape[lef_new_idx + 1: new_idx + 1])
                rechunk_nsplists[old_idx] = TensorReshape._reduce_nsplit_by_expand(nsplits[old_idx], chunk_expand)

                for i in range(lef_new_idx + 1, new_idx + 1):
                    reshape_nsplists[i] = (new_shape[i],)
                reshape_nsplists[lef_new_idx] = tuple(c // chunk_expand for c in rechunk_nsplists[old_idx])

                old_idx -= 1
                new_idx = lef_new_idx - 1

        assert np.prod([len(s) for s in rechunk_nsplists]) == \
               np.prod([len(s) for s in reshape_nsplists])
        return rechunk_nsplists, reshape_nsplists

    @staticmethod
    def _expand_nsplit_by_reduce(splits, reduced):
        if reduced == 1:
            return splits

        out = []
        for s in splits:
            x = s
            part = max(x / reduced, 1)
            while x >= 2 * part:
                out.append(int(part))
                x -= int(part)
            if x:
                out.append(x)
        assert sum(splits) == sum(out)
        return tuple(out)

    @staticmethod
    def _reduce_nsplit_by_expand(splits, expand):
        assert sum(splits) % expand == 0

        out = []
        residual = 0
        for chunk in splits:
            chunk += residual
            div = chunk // expand
            residual = chunk % expand
            good = expand * div
            if good:
                out.append(good)
        return tuple(out)

    @classmethod
    def tile(cls, op):
        in_tensor = op.input
        tensor = op.outputs[0]

        try:
            rechunk_nsplits, reshape_nsplits = cls._gen_reshape_rechunk_nsplits(
                in_tensor.shape, tensor.shape, in_tensor.nsplits)
            rechunked_tensor = in_tensor.rechunk(rechunk_nsplits).single_tiles()
            in_idxes = itertools.product(*[range(len(s)) for s in rechunk_nsplits])
            out_idxes = itertools.product(*[range(len(s)) for s in reshape_nsplits])
            out_shape = itertools.product(*[s for s in reshape_nsplits])
            out_chunks = []
            for input_idx, out_idx, out_shape in izip(in_idxes, out_idxes, out_shape):
                in_chunk = rechunked_tensor.cix[input_idx]
                chunk_op = op.copy().reset_key()
                chunk_op._newshape = out_shape
                out_chunk = chunk_op.new_chunk([in_chunk], out_shape, index=out_idx)
                out_chunks.append(out_chunk)

            new_op = op.copy()
            return new_op.new_tensors(op.inputs, tensor.shape,
                                      chunks=out_chunks, nsplits=reshape_nsplits)
        except ValueError:
            # shape incompatible, we will first do flatten, then reshape to the new shape
            # TODO(jisheng): try out shuffle if possible
            return [in_tensor.reshape(-1).single_tiles().reshape(tensor.shape).single_tiles()]


def reshape(a, newshape):
    """
    Gives a new shape to a tensor without changing its data.

    Parameters
    ----------
    a : array_like
        Tensor to be reshaped.
    newshape : int or tuple of ints
        The new shape should be compatible with the original shape. If
        an integer, then the result will be a 1-D tensor of that length.
        One shape dimension can be -1. In this case, the value is
        inferred from the length of the tensor and remaining dimensions.

    Returns
    -------
    reshaped_array : Tensor
        This will be a new view object if possible; otherwise, it will
        be a copy.

    See Also
    --------
    Tensor.reshape : Equivalent method.

    Notes
    -----
    It is not always possible to change the shape of a tensor without
    copying the data. If you want an error to be raised when the data is copied,
    you should assign the new shape to the shape attribute of the array::

    >>> import mars.tensor as mt

    >>> a = mt.arange(6).reshape((3, 2))
    >>> a.execute()
    array([[0, 1],
           [2, 3],
           [4, 5]])

    You can think of reshaping as first raveling the tensor (using the given
    index order), then inserting the elements from the raveled tensor into the
    new tensor using the same kind of index ordering as was used for the
    raveling.

    >>> mt.reshape(a, (2, 3)).execute()
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> mt.reshape(mt.ravel(a), (2, 3)).execute()
    array([[0, 1, 2],
           [3, 4, 5]])

    Examples
    --------
    >>> a = mt.array([[1,2,3], [4,5,6]])
    >>> mt.reshape(a, 6).execute()
    array([1, 2, 3, 4, 5, 6])

    >>> mt.reshape(a, (3,-1)).execute()       # the unspecified value is inferred to be 2
    array([[1, 2],
           [3, 4],
           [5, 6]])
    """
    a = astensor(a)
    if isinstance(newshape, six.integer_types):
        newshape = (newshape,)
    else:
        newshape = tuple(int(s) for s in newshape)

    if np.isnan(sum(a.shape)):
        raise ValueError('tensor shape is unknown, {0}'.format(a.shape))

    known_shape = [s for s in newshape if s >= 0]
    missing_dim = len(newshape) - len(known_shape)
    if missing_dim > 1:
        raise ValueError('can only specify one unknown dimension')
    if missing_dim == 1:
        known_size = np.prod(known_shape)
        newshape = tuple(int(a.size / known_size) if s < 0 and known_size > 0 else s
                         for s in newshape)

    if a.size != np.prod(newshape):
        raise ValueError('cannot reshape array of size {0} into shape {1}'.format(a.size, newshape))

    if a.shape == newshape:
        # does not need to reshape
        return a

    op = TensorReshape(newshape, dtype=a.dtype)
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MarsError(RuntimeError):
    """
    """
    def __init__(self, msg=None):
        super(MarsError, self).__init__(msg)

    def __str__(self):
        if hasattr(self, 'message'):
            message = self.message
        else:
            message = self.args[0]  # py3
        return message or ''


class ResourceInsufficient(MarsError):
    def __init__(self, msg=None, **kwargs):
        super(ResourceInsufficient, self).__init__(msg)
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, item):
        return None


class StartArgumentError(MarsError):
    pass


class StorageExhausted(MarsError):
    pass


class WorkerDead(MarsError):
    pass


class DependencyMissing(MarsError):
    pass


class StoreFull(MarsError):
    pass


class StoreKeyExists(MarsError):
    pass


class ChecksumMismatch(MarsError):
    pass


class ExecutionInterrupted(MarsError):
    pass


class ExecutionFailed(MarsError):
    pass


class ExecutionNotStopped(MarsError):
    pass


class ExecutionStateUnknown(MarsError):
    pass


class ResponseMalformed(MarsError):
    pass


class SpillNotConfigured(MarsError):
    pass


class GraphNotExists(MarsError):
    pass


class PromiseTimeout(MarsError):
    pass


class SpillExhausted(MarsError):
    pass


class PinChunkFailed(MarsError):
    pass


class ObjectNotInPlasma(MarsError):
    pass


class WorkerProcessStopped(MarsError):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
from operator import attrgetter, mul
from weakref import WeakKeyDictionary, ref
from collections import Iterable
import threading

import numpy as np

from ..core import Entity
from ..compat import builtins, reduce
from ..graph import DAG
from ..tiles import Tilesable, handler
from ..serialize import SerializableWithKey, ValueType, ProviderType, \
    ListField, TupleField, DictField, DataTypeField, KeyField, BoolField
from ..utils import AttributeDict, on_serialize_shape, on_deserialize_shape, tokenize
from .expressions.utils import get_chunk_slices


class ChunkData(SerializableWithKey):
    __slots__ = '__weakref__',

    # required fields
    _shape = TupleField('shape', ValueType.int64,
                        on_serialize=on_serialize_shape, on_deserialize=on_deserialize_shape)
    _op = KeyField('op')  # store key of operand here
    # optional fields
    _dtype = DataTypeField('dtype')
    _index = TupleField('index', ValueType.uint32)
    _cached = BoolField('cached')
    _composed = ListField('composed', ValueType.reference('self'))
    _params = DictField('params', key_type=ValueType.string, on_deserialize=AttributeDict)

    def __init__(self, *args, **kwargs):
        extras = AttributeDict((k, kwargs.pop(k)) for k in set(kwargs) - set(self.__slots__))
        kwargs['_params'] = kwargs.pop('_params', extras)
        super(ChunkData, self).__init__(*args, **kwargs)

    def __repr__(self):
        return 'Chunk <op={0}, key={1}>'.format(self.op.__class__.__name__, self.key)

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from ..serialize.protos.chunk_pb2 import ChunkDef
            return ChunkDef
        return super(ChunkData, cls).cls(provider)

    @property
    def shape(self):
        return getattr(self, '_shape', None)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def index(self):
        return getattr(self, '_index', None)

    @property
    def op(self):
        try:
            return self._op
        except AttributeError:
            return None

    @property
    def cached(self):
        return getattr(self, '_cached', None)

    @property
    def inputs(self):
        return self.op.inputs

    @inputs.setter
    def inputs(self, new_inputs):
        self.op.inputs = new_inputs

    @property
    def dtype(self):
        return getattr(self, '_dtype', None) or self.op.dtype

    @property
    def nbytes(self):
        return np.prod(self.shape) * self.dtype.itemsize

    @property
    def composed(self):
        return getattr(self, '_composed', None)

    @property
    def device(self):
        return self.op.device

    def is_sparse(self):
        return self.op.is_sparse()

    issparse = is_sparse

    def update_key(self):
        object.__setattr__(self, '_key', tokenize(
            type(self), *(getattr(self, k, None) for k in self.__slots__ if k != '_index')))


class Chunk(Entity):
    __slots__ = ()
    _allow_data_type_ = (ChunkData,)


class TensorData(SerializableWithKey, Tilesable):
    __slots__ = '__weakref__', '_siblings'

    # required fields
    _shape = TupleField('shape', ValueType.int64,
                        on_serialize=on_serialize_shape, on_deserialize=on_deserialize_shape)
    _dtype = DataTypeField('dtype')
    _op = KeyField('op')
    # optional fields
    # `nsplits` means the sizes of chunks for each dimension
    _nsplits = TupleField('nsplits', ValueType.tuple(ValueType.uint64))
    _chunks = ListField('chunks', ValueType.reference(Chunk))
    _params = DictField('params', key_type=ValueType.string, on_deserialize=AttributeDict)

    def __init__(self, *args, **kwargs):
        extras = AttributeDict((k, kwargs.pop(k)) for k in set(kwargs) - set(self.__slots__))
        kwargs['_params'] = kwargs.pop('_params', extras)
        if '_nsplits' in kwargs:
            kwargs['_nsplits'] = tuple(tuple(s) for s in kwargs['_nsplits'])

        super(TensorData, self).__init__(*args, **kwargs)

        if hasattr(self, '_chunks') and self._chunks:
            self._chunks = sorted(self._chunks, key=attrgetter('index'))

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from ..serialize.protos.tensor_pb2 import TensorDef
            return TensorDef
        return super(TensorData, cls).cls(provider)

    def __repr__(self):
        return 'Tensor <op={0}, key={1}>'.format(self.op.__class__.__name__, self.key)

    @property
    def shape(self):
        if hasattr(self, '_shape') and self._shape is not None:
            return self._shape
        if hasattr(self, '_nsplits') and self._nsplits is not None:
            self._shape = tuple(builtins.sum(nsplit) for nsplit in self._nsplits)
            return self._shape

    @property
    def ndim(self):
        return len(self.shape)

    def __len__(self):
        try:
            return self.shape[0]
        except IndexError:
            if build_mode().is_build_mode:
                return 0
            raise TypeError('len() of unsized object')

    @property
    def nbytes(self):
        return np.prod(self.shape) * self.dtype.itemsize

    @property
    def chunk_shape(self):
        if hasattr(self, '_nsplits') and self._nsplits is not None:
            return tuple(map(len, self._nsplits))

    @property
    def chunks(self):
        return getattr(self, '_chunks', None)

    @property
    def op(self):
        return getattr(self, '_op', None)

    @property
    def nsplits(self):
        return getattr(self, '_nsplits', None)

    @property
    def size(self):
        return np.prod(self.shape).item()

    @property
    def inputs(self):
        return self.op.inputs or []

    @inputs.setter
    def inputs(self, new_inputs):
        self.op.inputs = new_inputs

    @property
    def dtype(self):
        return getattr(self, '_dtype', None) or self.op.dtype

    @property
    def params(self):
        return self._params

    @property
    def real(self):
        from .expressions.arithmetic import real
        return real(self)

    @property
    def imag(self):
        from .expressions.arithmetic import imag
        return imag(self)

    def get_chunk_slices(self, idx):
        return get_chunk_slices(self.nsplits, idx)

    def is_coarse(self):
        return not hasattr(self, '_chunks') or self._chunks is None or len(self._chunks) == 0

    def is_scalar(self):
        return self.ndim == 0

    isscalar = is_scalar

    def is_sparse(self):
        return self.op.is_sparse()

    issparse = is_sparse

    def tosparse(self):
        if self.issparse():
            return self

        from .expressions.datasource import fromdense
        return fromdense(self)

    def todense(self):
        if not self.issparse():
            return self

        from .expressions.datasource import fromsparse
        return fromsparse(self)

    @property
    def cix(self):
        if self.ndim == 0:
            return ChunksIndexer(self)

        try:
            if self not in _tensor_to_chunk_indexer:
                _tensor_to_chunk_indexer[self] = ChunksIndexer(self)

            return _tensor_to_chunk_indexer[self]
        except (TypeError, ValueError):
            return ChunksIndexer(self)

    def tiles(self):
        return handler.tiles(self)

    def single_tiles(self):
        return handler.single_tiles(self)

    def build_graph(self, graph=None, cls=DAG, tiled=False, compose=True):
        if tiled and self.is_coarse():
            self.tiles()

        graph = graph if graph is not None else cls()
        keys = None

        if tiled:
            nodes = list(c.data for c in self.chunks)
            keys = list(c.key for c in self.chunks)
        else:
            nodes = list(self.op.outputs)
        visited = set()
        while len(nodes) > 0:
            chunk = nodes.pop()
            visited.add(chunk)
            if not graph.contains(chunk):
                graph.add_node(chunk)
            children = chunk.inputs or []
            for c in children:
                if not graph.contains(c):
                    graph.add_node(c)
                if not graph.has_successor(c, chunk):
                    graph.add_edge(c, chunk)
            nodes.extend([c for c in itertools.chain(*[inp.op.outputs for inp in chunk.inputs or []])
                          if c not in visited])
        if tiled and compose:
            graph.compose(keys=keys)
        return graph

    def transpose(self, *axes):
        """
        Returns a view of the tensor with axes transposed.

        For a 1-D tensor, this has no effect. (To change between column and
        row vectors, first cast the 1-D tensor into a matrix object.)
        For a 2-D tensor, this is the usual matrix transpose.
        For an n-D tensor, if axes are given, their order indicates how the
        axes are permuted (see Examples). If axes are not provided and
        ``a.shape = (i[0], i[1], ... i[n-2], i[n-1])``, then
        ``a.transpose().shape = (i[n-1], i[n-2], ... i[1], i[0])``.

        Parameters
        ----------
        axes : None, tuple of ints, or `n` ints

         * None or no argument: reverses the order of the axes.

         * tuple of ints: `i` in the `j`-th place in the tuple means `a`'s
           `i`-th axis becomes `a.transpose()`'s `j`-th axis.

         * `n` ints: same as an n-tuple of the same ints (this form is
           intended simply as a "convenience" alternative to the tuple form)

        Returns
        -------
        out : Tensor
            View of `a`, with axes suitably permuted.

        See Also
        --------
        Tensor.T : Tensor property returning the tensor transposed.

        Examples
        --------
        >>> import mars.tensor as mt

        >>> a = mt.array([[1, 2], [3, 4]])
        >>> a.execute()
        array([[1, 2],
               [3, 4]])
        >>> a.transpose().execute()
        array([[1, 3],
               [2, 4]])
        >>> a.transpose((1, 0))
        array([[1, 3],
               [2, 4]])
        >>> a.transpose(1, 0).execute()
        array([[1, 3],
               [2, 4]])
        """
        from .expressions.base import transpose

        if len(axes) == 1 and isinstance(axes[0], Iterable):
            axes = axes[0]

        return transpose(self, axes)

    @property
    def T(self):
        """
        Same as self.transpose(), except that self is returned if
        self.ndim < 2.

        Examples
        --------
        >>> import mars.tensor as mt

        >>> x = mt.array([[1.,2.],[3.,4.]])
        >>> x.execute()
        array([[ 1.,  2.],
               [ 3.,  4.]])
        >>> x.T.execute()
        array([[ 1.,  3.],
               [ 2.,  4.]])
        >>> x = mt.array([1.,2.,3.,4.])
        >>> x.execute()
        array([ 1.,  2.,  3.,  4.])
        >>> x.T.execute()
        array([ 1.,  2.,  3.,  4.])
        """
        return self.transpose()

    def reshape(self, shape, *shapes):
        """
        Returns a tensor containing the same data with a new shape.

        Refer to `mt.reshape` for full documentation.

        See Also
        --------
        mt.reshape : equivalent function

        Notes
        -----
        Unlike the free function `mt.reshape`, this method on `Tensor` allows
        the elements of the shape parameter to be passed in as separate arguments.
        For example, ``a.reshape(10, 11)`` is equivalent to
        ``a.reshape((10, 11))``.
        """
        from .expressions.reshape import reshape

        if isinstance(shape, Iterable):
            shape = tuple(shape)
        else:
            shape = (shape,)
        shape += shapes

        return reshape(self, shape)

    def ravel(self):
        """
        Return a flattened tensor.

        Refer to `mt.ravel` for full documentation.

        See Also
        --------
        mt.ravel : equivalent function
        """
        from .expressions.base import ravel

        return ravel(self)

    flatten = ravel

    def _equals(self, o):
        return self is o

    def execute(self, session=None, **kw):
        from ..session import Session

        if session is None:
            session = Session.default_or_local()
        return session.run(self, **kw)

    def _set_execute_session(self, session):
        _cleaner.register(self, session)

    _execute_session = property(fset=_set_execute_session)

    def visualize(self, graph_attrs=None, node_attrs=None, **kw):
        from graphviz import Source

        g = self.build_graph(**kw)
        dot = g.to_dot(graph_attrs=graph_attrs, node_attrs=node_attrs)

        return Source(dot)


class ExecutableTuple(tuple):
    def execute(self, session=None, **kw):
        from ..session import Session

        if session is None:
            session = Session.default_or_local()
        return session.run(*self, **kw)


class ChunksIndexer(object):
    __slots__ = '_tensor',

    def __init__(self, tensor):
        self._tensor = tensor

    def __getitem__(self, item):
        if isinstance(item, tuple):
            if len(item) == 0 and self._tensor.is_scalar():
                return self._tensor.chunks[0]
            elif all(np.issubdtype(type(it), np.integer) for it in item):
                if len(item) != self._tensor.ndim:
                    raise ValueError('Cannot get tensor chunk by %s, expect length %d' % (
                        item, self._tensor.ndim))

                s = self._tensor.chunk_shape
                item = tuple(i if i >= 0 else i + s for i, s in zip(item, s))
                idx = sum(idx * reduce(mul, s[i+1:], 1) for i, idx
                          in zip(itertools.count(0), item))
                return self._tensor._chunks[idx]

        raise ValueError('Cannot get tensor chunk by {0}'.format(item))


class Tensor(Entity):
    __slots__ = ()
    _allow_data_type_ = (TensorData,)

    def __len__(self):
        return len(self._data)

    def copy(self):
        return Tensor(self._data)

    def tiles(self):
        return handler.tiles(self)

    def single_tiles(self):
        return handler.single_tiles(self)

    @property
    def shape(self):
        return self.data.shape

    @shape.setter
    def shape(self, new_shape):
        self._data = self._data.reshape(new_shape).data

    @property
    def real(self):
        return self.data.real

    @real.setter
    def real(self, new_real):
        from .expressions.arithmetic.setreal import set_real

        self._data = set_real(self._data, new_real).data

    @property
    def imag(self):
        return self.data.imag

    @imag.setter
    def imag(self, new_imag):
        from .expressions.arithmetic.setimag import set_imag

        self._data = set_imag(self._data, new_imag).data


class SparseTensor(Tensor):
    __slots__ = ()


TENSOR_TYPE = (Tensor, TensorData)
CHUNK_TYPE = (Chunk, ChunkData)

_tensor_to_chunk_indexer = WeakKeyDictionary()
_threading_local = threading.local()


class _TensorSession(object):
    def __init__(self, tensor, session):
        key = tensor.key, tensor.id

        def cb(_, sess=ref(session)):
            s = sess()
            if s:
                s.decref(key)
        self._tensor = ref(tensor, cb)


class _TensorCleaner(object):
    def __init__(self):
        self._tensor_to_sessions = WeakKeyDictionary()

    def register(self, tensor, session):
        with build_mode():
            self._tensor_to_sessions[tensor] = _TensorSession(tensor, session)


# we don't use __del__ to decref because a tensor holds an op,
# and op's outputs contains the tensor, so a circular references exists
_cleaner = _TensorCleaner()


class BuildMode(object):
    def __init__(self):
        self.is_build_mode = False
        self._old_mode = None

    def __enter__(self):
        if self._old_mode is None:
            # check to prevent nested enter and exit
            self._old_mode = self.is_build_mode
            self.is_build_mode = True

    def __exit__(self, *_):
        if self._old_mode is not None:
            self.is_build_mode = self._old_mode
            self._old_mode = None


def build_mode():
    ret = getattr(_threading_local, 'build_mode', None)
    if ret is None:
        ret = BuildMode()
        _threading_local.build_mode = ret

    return ret
